"""Global definition of known notification types."""

from typing import List, Set

from django.db import models
from django.utils.translation import gettext_lazy as _

from structures.app_settings import STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS


class NotificationType(models.TextChoices):
    """Definition of all supported notification types."""

    # upwell structures
    OWNERSHIP_TRANSFERRED = "OwnershipTransferred", _(
        "Upwell structure ownership transferred"
    )
    STRUCTURE_ANCHORING = "StructureAnchoring", _("Upwell structure anchoring")
    STRUCTURE_DESTROYED = "StructureDestroyed", _("Upwell structure destroyed")
    STRUCTURE_FUEL_ALERT = "StructureFuelAlert", _("Upwell structure fuel alert")
    STRUCTURE_JUMP_FUEL_ALERT = "StructureJumpFuelAlert", _(
        "Upwell structure jump fuel alert"
    )
    STRUCTURE_LOST_ARMOR = "StructureLostArmor", _("Upwell structure lost armor")
    STRUCTURE_LOST_SHIELD = "StructureLostShields", _("Upwell structure lost shields")
    STRUCTURE_LOW_REAGENTS_ALERT = "StructureLowReagentsAlert", _(
        "Structure low reagents alert"
    )
    STRUCTURE_NO_REAGENTS_ALERT = "StructureNoReagentsAlert", _(
        "Structure no reagents alert"
    )
    STRUCTURE_ONLINE = "StructureOnline", _("Upwell structure went online")
    STRUCTURE_REFUELED_EXTRA = "StructureRefueledExtra", _("Upwell structure refueled")
    STRUCTURE_REINFORCEMENT_CHANGED = "StructuresReinforcementChanged", _(
        "Upwell structure reinforcement time changed"
    )
    STRUCTURE_SERVICES_OFFLINE = "StructureServicesOffline", _(
        "Upwell structure services went offline"
    )
    STRUCTURE_UNANCHORING = "StructureUnanchoring", _("Upwell structure unanchoring")
    STRUCTURE_UNDER_ATTACK = "StructureUnderAttack", _(
        "Upwell structure is under attack"
    )
    STRUCTURE_WENT_HIGH_POWER = "StructureWentHighPower", _(
        "Upwell structure went high power"
    )
    STRUCTURE_WENT_LOW_POWER = "StructureWentLowPower", _(
        "Upwell structure went low power"
    )

    # Skyhook structures
    SKYHOOK_DEPLOYED = "SkyhookDeployed", _("Skyhook deployed")
    SKYHOOK_DESTROYED = "SkyhookDestroyed", _("Skyhook destroyed")
    SKYHOOK_LOST_SHIELDS = "SkyhookLostShields", _("Skyhook lost shields")
    SKYHOOK_ONLINE = "SkyhookOnline", _("Skyhook online")
    SKYHOOK_UNDER_ATTACK = "SkyhookUnderAttack", _("Skyhook under attack")

    # customs offices
    ORBITAL_ATTACKED = "OrbitalAttacked", _("Customs office attacked")
    ORBITAL_REINFORCED = "OrbitalReinforced", _("Customs office reinforced")

    # starbases
    TOWER_ALERT_MSG = "TowerAlertMsg", _("Starbase attacked")
    TOWER_REFUELED_EXTRA = "TowerRefueledExtra", _("Starbase refueled (BETA)")
    TOWER_REINFORCED_EXTRA = "TowerReinforcedExtra", _("Starbase reinforced (BETA)")
    TOWER_RESOURCE_ALERT_MSG = "TowerResourceAlertMsg", _("Starbase fuel alert")

    # moon mining
    MOONMINING_AUTOMATIC_FRACTURE = "MoonminingAutomaticFracture", _(
        "Moon mining automatic fracture triggered"
    )
    MOONMINING_EXTRACTION_CANCELLED = "MoonminingExtractionCancelled", _(
        "Moon mining extraction cancelled"
    )
    MOONMINING_EXTRACTION_FINISHED = "MoonminingExtractionFinished", _(
        "Moon mining extraction finished"
    )
    MOONMINING_EXTRACTION_STARTED = "MoonminingExtractionStarted", _(
        "Moon mining extraction started"
    )
    MOONMINING_LASER_FIRED = "MoonminingLaserFired", _("Moonmining laser fired")

    # sov
    SOV_ALL_CLAIM_ACQUIRED_MSG = "SovAllClaimAquiredMsg", _(
        "Sovereignty claim acknowledgment"  # SovAllClaimAquiredMsg [sic!]
    )
    SOV_ALL_CLAIM_LOST_MSG = "SovAllClaimLostMsg", _("Sovereignty lost")
    SOV_ALL_ANCHORING_MSG = "AllAnchoringMsg", _(
        "Structure anchoring in alliance space"
    )
    SOV_ENTOSIS_CAPTURE_STARTED = "EntosisCaptureStarted", _(
        "Sovereignty entosis capture started"
    )
    SOV_STRUCTURE_DESTROYED = "SovStructureDestroyed", _(
        "Sovereignty structure destroyed"
    )
    SOV_STRUCTURE_REINFORCED = "SovStructureReinforced", _(
        "Sovereignty structure reinforced"
    )
    SOV_COMMAND_NODE_EVENT_STARTED = "SovCommandNodeEventStarted", _(
        "Sovereignty command node event started"
    )

    # wars
    WAR_ACCEPTED_ALLY = "AcceptedAlly", _("War accepted ally")
    WAR_ALL_WAR_CORP_JOINED_ALLIANCE_MSG = "AllWarCorpJoinedAllianceMsg", _(
        "Alliance war corporation joined alliance message"
    )
    WAR_ALL_WAR_SURRENDER_MSG = "AllWarSurrenderMsg", _(
        "Alliance war surrender message"
    )
    WAR_ALLY_JOINED_WAR_AGGRESSOR_MSG = "AllyJoinedWarAggressorMsg", _(
        "War ally joined aggressor"
    )
    WAR_ALLY_JOINED_WAR_ALLY_MSG = "AllyJoinedWarAllyMsg", _("War ally joined ally")
    WAR_ALLY_JOINED_WAR_DEFENDER_MSG = "AllyJoinedWarDefenderMsg", _(
        "War ally joined defender"
    )
    WAR_CORP_WAR_SURRENDER_MSG = "CorpWarSurrenderMsg", _("War party surrendered")
    WAR_CORPORATION_BECAME_ELIGIBLE = "CorpBecameWarEligible", _(
        "War corporation became eligible"
    )
    WAR_CORPORATION_NO_LONGER_ELIGIBLE = "CorpNoLongerWarEligible", _(
        "War corporation no longer eligible"
    )
    WAR_DECLARE_WAR = "DeclareWar", _("War declared")
    WAR_HQ_REMOVED_FROM_SPACE = "WarHQRemovedFromSpace", _("War HQ removed from space")
    WAR_INVALID = "WarInvalid", _("War invalid")
    WAR_MERC_OFFERED_NEGOTIATION_MSG = "MercOfferedNegotiationMsg", _(
        "War mercenary offered negotiation message"
    )
    WAR_MERC_OFFER_RETRACTED_MSG = "MercOfferRetractedMsg", _(
        "War mercenary offer retracted message"
    )
    WAR_OFFERED_SURRENDER = "OfferedSurrender", _("War offered surrender")
    WAR_OFFERED_TO_ALLY = "OfferedToAlly", _("War offered to become ally")
    WAR_WAR_ADOPTED = "WarAdopted", _("War adopted")  # FIXME: Should be "WarAdopted "
    WAR_WAR_DECLARED = "WarDeclared", _("War declared")
    WAR_WAR_INHERITED = "WarInherited", _("War inherited")
    WAR_WAR_RETRACTED_BY_CONCORD = "WarRetractedByConcord", _(
        "War retracted by Concord"
    )
    WAR_WAR_SURRENDER_OFFER_MSG = "WarSurrenderOfferMsg", _("War surrender offered")

    # corporation membership
    CHAR_APP_ACCEPT_MSG = "CharAppAcceptMsg", _("Character joins corporation")
    CORP_APP_INVITED_MSG = "CorpAppInvitedMsg", _(
        "Character invited to join corporation"
    )
    CORP_APP_NEW_MSG = "CorpAppNewMsg", _("Character submitted application")
    CORP_APP_REJECT_MSG = "CharAppRejectMsg", _("Corp application rejected message")
    CORP_APP_REJECT_CUSTOM_MSG = "CorpAppRejectCustomMsg", _(
        "Corp application rejected custom message"
    )
    CHAR_APP_WITHDRAW_MSG = "CharAppWithdrawMsg", _("Character withdrew application")
    CHAR_LEFT_CORP_MSG = "CharLeftCorpMsg", _("Character leaves corporation")

    # billing
    BILLING_BILL_OUT_OF_MONEY_MSG = "BillOutOfMoneyMsg", _("Bill out of money")
    BILLING_CORP_ALL_BILL_MSG = "CorpAllBillMsg", _("Corp alliance billing message")
    BILLING_I_HUB_BILL_ABOUT_TO_EXPIRE = (
        "InfrastructureHubBillAboutToExpire",
        _("I-HUB bill about to expire"),
    )
    BILLING_I_HUB_DESTROYED_BY_BILL_FAILURE = (
        "IHubDestroyedByBillFailure",
        _("I_HUB destroyed by bill failure"),
    )

    @classmethod
    def esi_notifications(cls) -> Set["NotificationType"]:
        """Return all ESI notification types."""
        return set(cls) - cls.generated_notifications()

    @classmethod
    def generated_notifications(cls) -> Set["NotificationType"]:
        """Return all generated notification types."""
        return {
            cls.STRUCTURE_JUMP_FUEL_ALERT,
            cls.STRUCTURE_REFUELED_EXTRA,
            cls.TOWER_REFUELED_EXTRA,
            cls.TOWER_REINFORCED_EXTRA,
        }

    @classmethod
    def webhook_defaults(cls) -> List["NotificationType"]:
        """List of default notifications for new webhooks."""
        return [
            cls.ORBITAL_ATTACKED,
            cls.ORBITAL_REINFORCED,
            cls.SKYHOOK_DESTROYED,
            cls.SKYHOOK_LOST_SHIELDS,
            cls.SKYHOOK_ONLINE,
            cls.SKYHOOK_UNDER_ATTACK,
            cls.SOV_STRUCTURE_DESTROYED,
            cls.SOV_STRUCTURE_REINFORCED,
            cls.STRUCTURE_ANCHORING,
            cls.STRUCTURE_DESTROYED,
            cls.STRUCTURE_FUEL_ALERT,
            cls.STRUCTURE_LOST_ARMOR,
            cls.STRUCTURE_LOST_SHIELD,
            cls.STRUCTURE_LOW_REAGENTS_ALERT,
            cls.STRUCTURE_NO_REAGENTS_ALERT,
            cls.STRUCTURE_ONLINE,
            cls.STRUCTURE_SERVICES_OFFLINE,
            cls.STRUCTURE_UNDER_ATTACK,
            cls.STRUCTURE_WENT_HIGH_POWER,
            cls.STRUCTURE_WENT_LOW_POWER,
            cls.TOWER_ALERT_MSG,
            cls.TOWER_RESOURCE_ALERT_MSG,
        ]

    @classmethod
    def relevant_for_timerboard(cls) -> Set["NotificationType"]:
        """Notification types that can create timers."""
        return {
            cls.MOONMINING_EXTRACTION_CANCELLED,
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.ORBITAL_REINFORCED,
            cls.SOV_STRUCTURE_REINFORCED,
            cls.STRUCTURE_LOST_ARMOR,
            cls.STRUCTURE_LOST_SHIELD,
            cls.TOWER_REINFORCED_EXTRA,
            cls.SKYHOOK_LOST_SHIELDS,
        }

    @classmethod
    def relevant_for_alliance_level(cls) -> Set["NotificationType"]:
        """Notification types that require the alliance level flag."""
        return {
            # billing
            cls.BILLING_BILL_OUT_OF_MONEY_MSG,
            cls.BILLING_I_HUB_BILL_ABOUT_TO_EXPIRE,
            cls.BILLING_I_HUB_DESTROYED_BY_BILL_FAILURE,
            # sov
            cls.SOV_ALL_CLAIM_ACQUIRED_MSG,
            cls.SOV_ALL_CLAIM_LOST_MSG,
            cls.SOV_COMMAND_NODE_EVENT_STARTED,
            cls.SOV_ENTOSIS_CAPTURE_STARTED,
            cls.SOV_STRUCTURE_DESTROYED,
            cls.SOV_STRUCTURE_REINFORCED,
            # cls.SOV_ALL_ANCHORING_MSG, # This notif is not broadcasted to all corporations
            # wars
            cls.WAR_ALLY_JOINED_WAR_AGGRESSOR_MSG,
            cls.WAR_ALLY_JOINED_WAR_ALLY_MSG,
            cls.WAR_ALLY_JOINED_WAR_DEFENDER_MSG,
            cls.WAR_CORP_WAR_SURRENDER_MSG,
            cls.WAR_CORPORATION_BECAME_ELIGIBLE,
            cls.WAR_CORPORATION_NO_LONGER_ELIGIBLE,
            cls.WAR_WAR_ADOPTED,
            cls.WAR_WAR_DECLARED,
            cls.WAR_WAR_INHERITED,
            cls.WAR_WAR_RETRACTED_BY_CONCORD,
            cls.WAR_WAR_SURRENDER_OFFER_MSG,
        }

    @classmethod
    def relevant_for_moonmining(cls) -> Set["NotificationType"]:
        """Notification types about moon mining."""
        return {
            cls.MOONMINING_AUTOMATIC_FRACTURE,
            cls.MOONMINING_EXTRACTION_CANCELLED,
            cls.MOONMINING_EXTRACTION_FINISHED,
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.MOONMINING_LASER_FIRED,
        }

    @classmethod
    def structure_related(cls) -> Set["NotificationType"]:
        """Notification types that are related to a structure."""
        return {
            cls.MOONMINING_AUTOMATIC_FRACTURE,
            cls.MOONMINING_EXTRACTION_CANCELLED,
            cls.MOONMINING_EXTRACTION_FINISHED,
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.MOONMINING_LASER_FIRED,
            cls.ORBITAL_ATTACKED,
            cls.ORBITAL_REINFORCED,
            cls.OWNERSHIP_TRANSFERRED,
            cls.STRUCTURE_ANCHORING,
            cls.STRUCTURE_DESTROYED,
            cls.STRUCTURE_FUEL_ALERT,
            cls.STRUCTURE_JUMP_FUEL_ALERT,
            cls.STRUCTURE_LOST_ARMOR,
            cls.STRUCTURE_LOST_SHIELD,
            cls.STRUCTURE_LOW_REAGENTS_ALERT,
            cls.STRUCTURE_NO_REAGENTS_ALERT,
            cls.STRUCTURE_ONLINE,
            cls.STRUCTURE_REFUELED_EXTRA,
            cls.STRUCTURE_REINFORCEMENT_CHANGED,
            cls.STRUCTURE_SERVICES_OFFLINE,
            cls.STRUCTURE_UNANCHORING,
            cls.STRUCTURE_UNDER_ATTACK,
            cls.STRUCTURE_WENT_HIGH_POWER,
            cls.STRUCTURE_WENT_LOW_POWER,
            cls.TOWER_ALERT_MSG,
            cls.TOWER_REFUELED_EXTRA,
            cls.TOWER_REINFORCED_EXTRA,
            cls.TOWER_RESOURCE_ALERT_MSG,
        }

    @classmethod
    def relevant_for_forwarding(cls) -> Set["NotificationType"]:
        """Notification types that are forwarded to Discord."""
        my_set = cls.values_enabled()
        # if STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS:
        #     my_set.discard(cls.STRUCTURE_FUEL_ALERT)
        #     my_set.discard(cls.TOWER_RESOURCE_ALERT_MSG)
        return my_set

    @classmethod
    def values_enabled(cls) -> Set["NotificationType"]:
        """Values of enabled notif types only."""
        my_set = set(cls.values)  # type: ignore
        if not STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS:
            my_set.discard(cls.STRUCTURE_REFUELED_EXTRA)
            my_set.discard(cls.TOWER_REFUELED_EXTRA)
        return my_set

    @classmethod
    def choices_enabled(cls) -> List[tuple]:
        """Choices list containing enabled notif types only."""
        return [choice for choice in cls.choices if choice[0] in cls.values_enabled()]



"""Settings for Structures."""

from app_utils.app_settings import clean_setting

STRUCTURES_ADD_TIMERS = clean_setting("STRUCTURES_ADD_TIMERS", True)
"""Whether to automatically add timers for certain notifications on the timerboard
(will have no effect if aa-timerboard app is not installed).

Will create timers from anchoring, lost shield and lost armor notifications.
"""

STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED = clean_setting(
    "STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", True
)
"""Whether admins will get notifications about import events
like when someone adds a structure owner and when services for an owner are down.
"""

STRUCTURES_DEFAULT_LANGUAGE = clean_setting("STRUCTURES_DEFAULT_LANGUAGE", "en")
"""Sets the default language to be used in case no language can be determined
e.g. this language will be used when creating timers
Please use the language codes as defined in the base.py settings file.
"""

STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED = clean_setting(
    "STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED", False
)
"""Whether the structure list has default tags filter enabled by default."""

STRUCTURES_DEVELOPER_MODE = clean_setting("STRUCTURES_DEVELOPER_MODE", False)
"""Enables features for developers e.g. write access to all models in admin
and writing raw data received from ESI.

:meta private:
"""

STRUCTURES_DEFAULT_PAGE_LENGTH = clean_setting("STRUCTURES_DEFAULT_PAGE_LENGTH", 10)
"""Default page size for structure list.
Must be an integer value from the current options as seen in the app.
"""

STRUCTURES_FEATURE_CUSTOMS_OFFICES = clean_setting(
    "STRUCTURES_FEATURE_CUSTOMS_OFFICES", True
)
"""Enable / disable custom offices feature."""

STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS = clean_setting(
    "STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True
)
"""Enable / disable refueled notifications feature."""

STRUCTURES_FEATURE_SKYHOOKS = clean_setting("STRUCTURES_FEATURE_SKYHOOKS", False)
"""Show skyhooks in structures list."""


STRUCTURES_FEATURE_STARBASES = clean_setting("STRUCTURES_FEATURE_STARBASES", True)
"""Enable / disable starbases feature."""

STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES = clean_setting(
    "STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES", 3
)
"""Max retries before a character is disabled when ESI claims the character
is not a director (Since this sometimes is reported wrongly by ESI).
"""

STRUCTURES_ESI_TIMEOUT_ENABLED = clean_setting("STRUCTURES_ESI_TIMEOUT_ENABLED", True)
"""Whether ESI timeout is enabled."""


STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION = clean_setting(
    "STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION", 24
)
"""Defines after how many hours a notification becomes stale.
Stale notification will no longer be sent automatically.
"""

STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED = clean_setting(
    "STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", True
)
"""Whether to create / remove timers from moon extraction notifications."""

STRUCTURES_NOTIFICATION_MAX_RETRIES = clean_setting(
    "STRUCTURES_NOTIFICATION_MAX_RETRIES", 3
)
"""Max number of retries for sending a notification
if an error occurred e.g. rate limiting.
"""

STRUCTURES_NOTIFICATION_SET_AVATAR = clean_setting(
    "STRUCTURES_NOTIFICATION_SET_AVATAR", True
)
"""Wether structures sets the name and avatar icon of a webhook.
When ``False`` the webhook will use it's own values as set on the platform.
"""

STRUCTURES_NOTIFICATION_SHOW_MOON_ORE = clean_setting(
    "STRUCTURES_NOTIFICATION_SHOW_MOON_ORE", True
)
"""Wether ore details are shown on moon timers."""


STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES = clean_setting(
    "STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 40
)
"""Max time in minutes since last successful notification sync
before service is reported as down.
"""

STRUCTURES_NOTIFICATION_WAIT_SEC = clean_setting("STRUCTURES_NOTIFICATION_WAIT_SEC", 5)
"""Default wait time in seconds before retrying to send a notification
to Discord after an error occurred.
"""

STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED = clean_setting(
    "STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED", False
)
"""Enables archiving of all notifications received from ESI to files
notifications will by stored into one continuous file per corporations.

:meta private:
"""

STRUCTURES_PAGING_ENABLED = clean_setting("STRUCTURES_PAGING_ENABLED", True)
"""Whether paging is enabled for the structure list"""

STRUCTURES_REPORT_NPC_ATTACKS = clean_setting("STRUCTURES_REPORT_NPC_ATTACKS", True)
"""Enable / disable sending notifications for attacks by NPCs
(structure reinforcements are still reported).
"""

STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE = clean_setting(
    "STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE", True
)
"""Whether fuel expires in structures browser is shown as absolute value."""

STRUCTURES_SHOW_JUMP_GATES = clean_setting("STRUCTURES_SHOW_JUMP_GATES", True)
"""Whether to show the jump gates tab."""

STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES = clean_setting(
    "STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 120
)
"""Max time in minutes since last successful structures sync
before service is reported as down.
"""

STRUCTURES_TASKS_TIME_LIMIT = clean_setting("STRUCTURES_TASKS_TIME_LIMIT", 7200)
"""Hard timeout for tasks in seconds to reduce task accumulation during outages."""

STRUCTURES_TIMERS_ARE_CORP_RESTRICTED = clean_setting(
    "STRUCTURES_TIMERS_ARE_CORP_RESTRICTED", False
)
"""Whether created timers are corp restricted on the timerboard."""

STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS = clean_setting(
    "STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS", False
)
"""This allows you to turn off ESI fuel alert notifications
to use the Structure's generated fuel notifications exclusively.
"""

# INTERNAL SETTINGS

# Number of notifications to count for short mean turnaround time
STRUCTURES_NOTIFICATION_TURNAROUND_SHORT = clean_setting(
    "STRUCTURES_NOTIFICATION_TURNAROUND_SHORT", 5
)

# Number of notifications to count for medium mean turnaround time
STRUCTURES_NOTIFICATION_TURNAROUND_MEDIUM = clean_setting(
    "STRUCTURES_NOTIFICATION_TURNAROUND_MEDIUM", 15
)

# Number of notifications to count for long mean turnaround time
STRUCTURES_NOTIFICATION_TURNAROUND_LONG = clean_setting(
    "STRUCTURES_NOTIFICATION_TURNAROUND_LONG", 50
)

# Turnaround duration with more than x seconds are regarded as outliers
# and will be ignored when calculating the averages
STRUCTURES_NOTIFICATION_TURNAROUND_MAX_VALID = clean_setting(
    "STRUCTURES_NOTIFICATION_TURNAROUND_MAX_VALID", 3600
)

# Timeout for throttled issue notifications to users and admins in seconds.
STRUCTURES_NOTIFY_THROTTLED_TIMEOUT = clean_setting(
    "STRUCTURES_NOTIFY_THROTTLED_TIMEOUT", 86400
)



import re
from unittest.mock import patch

from requests.exceptions import HTTPError

from django.contrib.auth.models import Group

from app_utils.django import app_labels
from app_utils.testing import NoSocketsTestCase

from structures.models import Notification
from structures.tests.testdata.factories import (
    OwnerFactory,
    StructureFactory,
    WebhookFactory,
)
from structures.tests.testdata.helpers import (
    clone_notification,
    load_eve_entities,
    load_notification_entities,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.notifications"

if "discord" in app_labels():

    @patch(MODULE_PATH + ".Notification._import_discord")
    @patch(MODULE_PATH + ".Webhook.send_message", spec=True)
    class TestGroupPings(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            load_eveuniverse()
            load_eve_entities()

            cls.group_1 = Group.objects.create(name="Dummy Group 1")
            cls.group_2 = Group.objects.create(name="Dummy Group 2")
            cls.owner = OwnerFactory()
            load_notification_entities(cls.owner)
            StructureFactory(id=1000000000001)

        @staticmethod
        def _my_group_to_role(group: Group) -> dict:
            if not isinstance(group, Group):
                raise TypeError("group must be of type Group")
            return {"id": group.pk, "name": group.name}

        def test_can_ping_via_webhook(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = WebhookFactory()
            webhook.ping_groups.add(self.group_1)
            obj = clone_notification(
                Notification.objects.get(notification_id=1000000509)
            )
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertIn(f"<@&{self.group_1.pk}>", kwargs["content"])

        def test_can_ping_via_owner(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = WebhookFactory()
            self.owner.ping_groups.add(self.group_2)
            obj = clone_notification(
                Notification.objects.get(notification_id=1000000509)
            )
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertIn(f"<@&{self.group_2.pk}>", kwargs["content"])

        def test_can_ping_both(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = WebhookFactory()
            webhook.ping_groups.add(self.group_1)
            self.owner.ping_groups.add(self.group_2)
            obj = clone_notification(
                Notification.objects.get(notification_id=1000000509)
            )
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertIn(f"<@&{self.group_1.pk}>", kwargs["content"])
            self.assertIn(f"<@&{self.group_2.pk}>", kwargs["content"])

        def test_no_ping_if_not_set(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = WebhookFactory()
            obj = clone_notification(
                Notification.objects.get(notification_id=1000000509)
            )
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertFalse(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertFalse(re.search(r"(<@&\d+>)", kwargs["content"]))

        def test_can_handle_http_error(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                HTTPError
            )
            webhook = WebhookFactory()
            webhook.ping_groups.add(self.group_1)
            obj = clone_notification(
                Notification.objects.get(notification_id=1000000509)
            )
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertFalse(re.search(r"(<@&\d+>)", kwargs["content"]))

        def test_should_abort_when_content_is_too_large(
            self, mock_send_message, mock_import_discord
        ):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = WebhookFactory()
            for i in range(286):
                group = Group.objects.create(name=f"Group {i+1}")
                webhook.ping_groups.add(group)
            obj = clone_notification(
                Notification.objects.get(notification_id=1000000509)
            )
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertFalse(result)



"""Views for statistics page."""

from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Count, F, Q
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from eveuniverse.core import dotlan, eveimageserver

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.views import link_html

from structures import __title__
from structures.constants import EveCategoryId, EveGroupId, EveTypeId
from structures.helpers import floating_icon_with_text_html
from structures.models import Structure

from .common import add_common_context, add_common_data_export

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def _default_if_none(value, default=None):
    """Return default if a value is None."""
    if value is None:
        return default
    return value


@login_required
@permission_required("structures.basic_access")
def statistics(request: HttpRequest) -> HttpResponse:
    """Return view to render Statistics page."""
    ajax_url = reverse("structures:structure_summary_data")
    data_export = add_common_data_export(
        {
            "ajax_url": ajax_url,
            "filter_titles": {"alliance": _("Alliance")},
        }
    )
    context = {"data_export": data_export}
    return render(request, "structures/statistics.html", add_common_context(context))


@login_required
@permission_required("structures.basic_access")
def structure_summary_data(request: HttpRequest) -> JsonResponse:
    """View returning data for structure summary page."""
    summary_qs = (
        Structure.objects.visible_for_user(request.user)
        .values(
            corporation_id=F("owner__corporation__corporation_id"),
            corporation_name=F("owner__corporation__corporation_name"),
            alliance_name=F("owner__corporation__alliance__alliance_name"),
            alliance_ticker=F("owner__corporation__alliance__alliance_ticker"),
        )
        .annotate(
            ec_count=Count(
                "id", filter=Q(eve_type__eve_group=EveGroupId.ENGINEERING_COMPLEX)
            )
        )
        .annotate(
            refinery_count=Count(
                "id", filter=Q(eve_type__eve_group=EveGroupId.REFINERY)
            )
        )
        .annotate(
            citadel_count=Count("id", filter=Q(eve_type__eve_group=EveGroupId.CITADEL))
        )
        .annotate(
            upwell_count=Count(
                "id",
                filter=Q(eve_type__eve_group__eve_category=EveCategoryId.STRUCTURE),
            )
        )
        .annotate(poco_count=Count("id", filter=Q(eve_type=EveTypeId.CUSTOMS_OFFICE)))
        .annotate(
            starbase_count=Count(
                "id", filter=Q(eve_type__eve_group__eve_category=EveCategoryId.STARBASE)
            )
        )
    )
    data = []
    for row in summary_qs:
        other_count = (
            row["upwell_count"]
            - row["ec_count"]
            - row["refinery_count"]
            - row["citadel_count"]
        )
        total = row["upwell_count"] + row["poco_count"] + row["starbase_count"]

        corporation_id = row["corporation_id"]
        corporation_name = row["corporation_name"]
        alliance_name = _default_if_none(row["alliance_name"], "")
        alliance_ticker = _default_if_none(row["alliance_ticker"], "")
        corporation_icon_url = eveimageserver.corporation_logo_url(
            corporation_id, size=64
        )
        owner_link = link_html(
            dotlan.corporation_url(corporation_name), corporation_name
        )
        owner_display_html = floating_icon_with_text_html(
            corporation_icon_url, [owner_link, alliance_ticker]
        )
        owner_html = {"display": owner_display_html, "value": corporation_name}
        alliance_name_str = (
            f"{alliance_name} [{alliance_ticker}]" if alliance_ticker else alliance_name
        )

        data.append(
            {
                "id": int(corporation_id),
                "owner": owner_html,
                "corporation_name": corporation_name,
                "alliance_name": alliance_name_str,
                "citadel_count": row["citadel_count"],
                "ec_count": row["ec_count"],
                "refinery_count": row["refinery_count"],
                "other_count": other_count,
                "poco_count": row["poco_count"],
                "starbase_count": row["starbase_count"],
                "total": total,
            }
        )
    return JsonResponse({"data": data})



"""Corporate embeds."""

# pylint: disable=missing-class-docstring

import dhooks_lite

from django.utils.translation import gettext as _

from structures.helpers import get_or_create_eve_entity
from structures.models import Notification, Webhook

from .helpers import (
    gen_corporation_link,
    gen_eve_entity_link,
    gen_eve_entity_link_from_id,
)
from .main import NotificationBaseEmbed


class NotificationCorpCharEmbed(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._character = get_or_create_eve_entity(id=self._data["charID"])
        self._corporation = get_or_create_eve_entity(id=self._data["corpID"])
        self._character_link = gen_eve_entity_link(self._character)
        self._corporation_link = gen_corporation_link(self._corporation.name)
        self._application_text = self._data.get("applicationText", "")
        self._thumbnail = dhooks_lite.Thumbnail(
            self._character.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationCorpAppNewMsg(NotificationCorpCharEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("New application from %(character_name)s") % {
            "character_name": self._character.name,
        }
        self._description = _(
            "New application from %(character_name)s to join %(corporation_name)s:\n"
            "> %(application_text)s"
            % {
                "character_name": self._character_link,
                "corporation_name": self._corporation_link,
                "application_text": self._application_text,
            }
        )
        self._color = Webhook.Color.INFO


class NotificationCorpAppInvitedMsg(NotificationCorpCharEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(character_name)s has been invited") % {
            "character_name": self._character.name
        }
        inviting_character = gen_eve_entity_link_from_id(
            self._data.get("invokingCharID")
        )
        self._description = _(
            "%(character_name)s has been invited to join %(corporation_name)s "
            "by %(inviting_character)s.\n"
            "Application:\n"
            "> %(application_text)s"
        ) % {
            "character_name": self._character_link,
            "corporation_name": self._corporation_link,
            "inviting_character": inviting_character,
            "application_text": self._application_text,
        }

        self._color = Webhook.Color.INFO


class NotificationCharAppRejectMsg(NotificationCorpCharEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(character_name)s rejects invitation") % {
            "character_name": self._character.name
        }
        self._description = _(
            "Application from %(character_name)s to join %(corporation_name)s "
            "has been rejected."
        ) % {
            "character_name": self._character_link,
            "corporation_name": self._corporation_link,
        }
        self._color = Webhook.Color.INFO


class NotificationCorpAppRejectCustomMsg(NotificationCorpCharEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Rejected application from %(character_name)s") % {
            "character_name": self._character.name
        }
        self._description = _(
            "Application from %(character_name)s to join %(corporation_name)s:\n"
            "> %(application_text)s\n"
            "Has been rejected:\n"
            "> %(customMessage)s"
        ) % {
            "character_name": self._character_link,
            "corporation_name": self._corporation_link,
            "application_text": self._application_text,
            "customMessage": self._data.get("customMessage", ""),
        }
        self._color = Webhook.Color.INFO


class NotificationCharAppWithdrawMsg(NotificationCorpCharEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(character_name)s withdrew his/her application") % {
            "character_name": self._character.name,
        }
        self._description = _(
            "%(character_name)s withdrew his/her application to join "
            "%(corporation_name)s:\n"
            "> %(application_text)s"
        ) % {
            "character_name": self._character_link,
            "corporation_name": self._corporation_link,
            "application_text": self._application_text,
        }

        self._color = Webhook.Color.INFO


class NotificationCharAppAcceptMsg(NotificationCorpCharEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(character_name)s joins %(corporation_name)s") % {
            "character_name": self._character.name,
            "corporation_name": self._corporation.name,
        }
        self._description = _(
            "%(character_name)s is now a member of %(corporation_name)s."
        ) % {
            "character_name": self._character_link,
            "corporation_name": self._corporation_link,
        }
        self._color = Webhook.Color.SUCCESS


class NotificationCharLeftCorpMsg(NotificationCorpCharEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(character_name)s has left %(corporation_name)s") % {
            "character_name": self._character.name,
            "corporation_name": self._corporation.name,
        }
        self._description = _(
            "%(character_name)s is no longer a member of %(corporation_name)s."
        ) % {
            "character_name": self._character_link,
            "corporation_name": self._corporation_link,
        }
        self._color = Webhook.Color.INFO



"""Notification related models for Structures."""

# pylint: disable = duplicate-code

import math
from typing import Optional, Tuple, Union

import dhooks_lite
import yaml
from multiselectfield import MultiSelectField
from multiselectfield.utils import get_max_length
from requests.exceptions import HTTPError

from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils import translation
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity, EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.django import app_labels
from app_utils.logging import LoggerAddTag
from app_utils.urls import static_file_absolute_url

from structures import __title__
from structures.app_settings import (  # STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS,
    STRUCTURES_ADD_TIMERS,
    STRUCTURES_DEFAULT_LANGUAGE,
    STRUCTURES_NOTIFICATION_SET_AVATAR,
    STRUCTURES_REPORT_NPC_ATTACKS,
)
from structures.constants import EveCategoryId, EveCorporationId, EveTypeId
from structures.core.notification_types import NotificationType
from structures.helpers import is_absolute_url
from structures.managers import (
    GeneratedNotificationManager,
    NotificationManager,
    WebhookManager,
)
from structures.webhooks.models import WebhookBase

from .structures_1 import Structure

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

# Supported languages
LANGUAGES = (
    ("en", _("English")),
    ("de", _("German")),
    ("es", _("Spanish")),
    ("zh-hans", _("Chinese Simplified")),
    ("ru", _("Russian")),
    ("ko", _("Korean")),
)


def get_default_notification_types():
    """DEPRECATED: generates a set of all existing notification types as default.
    Required to support older migrations."""
    return tuple(sorted([str(x[0]) for x in NotificationType.values]))


class Webhook(WebhookBase):
    """A destination for forwarding notification alerts."""

    notification_types = MultiSelectField(
        choices=NotificationType.choices,
        max_length=get_max_length(NotificationType.choices, None),
        default=NotificationType.webhook_defaults(),
        verbose_name=_("notification types"),
        help_text=_(
            "Select which type of notifications should be forwarded to this webhook"
        ),
    )
    language_code = models.CharField(
        max_length=8,
        choices=LANGUAGES,
        default=None,
        null=True,
        blank=True,
        verbose_name=_("language"),
        help_text=_("Language of notifications send to this webhook"),
    )
    is_default = models.BooleanField(
        default=False,
        verbose_name=_("is default"),
        help_text=_(
            "Whether owners have this webhook automatically pre-set when created"
        ),
    )
    has_default_pings_enabled = models.BooleanField(
        default=True,
        verbose_name=_("has default pings enabled"),
        help_text=(
            "To enable or disable pinging of notifications for this webhook "
            "e.g. with @everyone and @here"
        ),
    )
    ping_groups = models.ManyToManyField(
        Group,
        default=None,
        blank=True,
        related_name="+",
        verbose_name=_("ping groups"),
        help_text=_("Groups to be pinged for each notification - "),
    )
    objects = WebhookManager()

    class Meta:
        verbose_name = _("webhook")
        verbose_name_plural = _("webhooks")

    @staticmethod
    def text_bold(text) -> str:
        """Format the given text in bold."""
        return f"**{text}**" if text else ""


# pylint: disable = too-many-public-methods
class NotificationBase(models.Model):
    """Base model for a notification."""

    is_sent = models.BooleanField(
        default=False,
        verbose_name=_("is sent"),
        help_text=_("True when this notification has been forwarded to Discord"),
    )
    is_timer_added = models.BooleanField(
        null=True,
        default=False,
        verbose_name=_("is timer added"),
        help_text=_("True when a timer has been added for this notification"),
    )
    notif_type = models.CharField(
        max_length=100,
        default="",
        db_index=True,
        verbose_name=_("type"),
        help_text=_("Type of this notification"),
    )
    owner = models.ForeignKey(
        "Owner",
        on_delete=models.CASCADE,
        verbose_name=_("owner"),
        help_text=_("Corporation that owns this notification"),
    )
    structures = models.ManyToManyField(
        Structure,
        verbose_name=_("structures"),
        help_text=_("Structures this notification is about (if any)"),
    )

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._ping_type_override = None
        self._color_override = None
        self._parsed_text = {}

    def __str__(self) -> str:
        return f"{self.notification_id}:{self.notif_type}"

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"notification_id={self.notification_id}, "
            f"owner='{self.owner}', "
            f"notif_type='{self.notif_type}'"
            ")"
        )

    @property
    def is_alliance_level(self) -> bool:
        """Whether this is an alliance level notification."""
        return self.notif_type in NotificationType.relevant_for_alliance_level()

    @property
    def can_be_rendered(self) -> bool:
        """Whether this notification can be rendered in Discord."""
        return self.notif_type in NotificationType.values

    @property
    def can_have_timer(self) -> bool:
        """Whether this notification can have a timer."""
        return self.notif_type in NotificationType.relevant_for_timerboard()

    @property
    def is_structure_related(self) -> bool:
        """Return True if this notification is about a structure, else False."""
        return self.notif_type in NotificationType.structure_related()

    @property
    def is_generated(self) -> bool:
        """Return True if this notification is generated, else False."""
        raise NotImplementedError()

    @property
    def is_temporary(self) -> bool:
        """Return True if this is a temporary notification, else False."""
        raise NotImplementedError()

    @property
    def color_override(self) -> Optional[int]:
        """Return overridden color if set, else None."""
        return self._color_override

    @property
    def ping_type_override(self) -> Optional[Webhook.PingType]:
        """Return overridden ping type if set, else None."""
        return self._ping_type_override

    def is_npc_attacking(self) -> bool:
        """Return True if this notification is about NPC attacking, else False."""
        raise NotImplementedError()

    def parsed_text(self) -> dict:
        """Return parsed text of this notification."""
        return self._parsed_text

    def eve_moon(self, key: str = "moonID") -> EveMoon:
        """Return it's moon extracted from the notification text.
        Will raise KeyError if not found.
        """
        obj_id = self._parsed_text[key]
        obj, _ = EveMoon.objects.get_or_create_esi(id=obj_id)
        return obj

    def eve_planet(self, key: str = "planetID") -> EvePlanet:
        """Return it's moon extracted from the notification text.
        Will raise KeyError if not found.
        """
        obj_id = self._parsed_text[key]
        obj, _ = EvePlanet.objects.get_or_create_esi(id=obj_id)
        return obj

    def eve_solar_system(self, key: str = "solarSystemID") -> EveSolarSystem:
        """Return solar system extracted from the notification text.
        Will raise KeyError if not found.
        """
        obj_id = self._parsed_text[key]
        obj, _ = EveSolarSystem.objects.get_or_create_esi(id=obj_id)
        return obj

    def eve_structure_type(self, key: str = "structureTypeID") -> EveType:
        """Return structure type extracted from the notification text.
        Will raise KeyError if not found.
        """
        obj_id = self._parsed_text[key]
        obj, _ = EveType.objects.get_or_create_esi(id=obj_id)
        return obj

    def send_to_configured_webhooks(
        self,
        ping_type_override: Optional[Union[Webhook.PingType, str]] = None,
        use_color_override: bool = False,
        color_override: Optional[int] = None,
    ) -> Optional[bool]:
        """Send this notification to all active webhooks which have this
        notification type configured
        and apply filter for NPC attacks and alliance level if needed.

        Returns True, if notifications has been successfully send to webhooks
        Returns None, if owner has no fitting webhook
        Returns False, if sending to any webhooks failed
        """
        if self.filter_for_npc_attacks():
            logger.info("%s: Ignoring NPC attacks as configured", self)
            return None

        if self.filter_for_alliance_level():
            logger.debug(
                "%s: Alliance level notifications are not enabled for this owner", self
            )
            return None

        webhooks_qs = self.relevant_webhooks()
        if not webhooks_qs.exists():
            logger.warning("%s: No relevant webhook found", self)
            return None

        if ping_type_override:
            self._ping_type_override = Webhook.PingType(ping_type_override)
        if use_color_override:
            self._color_override = color_override
        success = True
        for webhook in webhooks_qs:
            success &= self.send_to_webhook(webhook)
        return success

    def filter_for_npc_attacks(self) -> bool:
        """True when notification to be filtered out due to npc attacks."""
        return not STRUCTURES_REPORT_NPC_ATTACKS and self.is_npc_attacking()

    def filter_for_alliance_level(self) -> bool:
        """True when notification to be filtered out due to alliance level."""
        return (
            self.is_alliance_level
            and self.owner.corporation.alliance is not None
            and not self.owner.is_alliance_main
        )

    def relevant_webhooks(self) -> models.QuerySet:
        """Determine relevant webhooks matching this notification type."""
        if not self.is_structure_related:
            structures_qs = Structure.objects.none()
        else:
            structures_qs = self.calc_related_structures()
        if (
            structures_qs.exists()
            and structures_qs.filter(webhooks__isnull=False).count() == 1
        ):
            webhooks_qs = structures_qs.first().webhooks.filter(
                notification_types__contains=self.notif_type, is_active=True
            )
        else:
            webhooks_qs = self.owner.webhooks.filter(
                notification_types__contains=self.notif_type, is_active=True
            )
        return webhooks_qs

    def calc_related_structures(self) -> models.QuerySet[Structure]:
        """Identify structures this notification is related to.

        Returns:
        - structures if any found or empty list if there are no related structures
        """
        if not self._parsed_text:
            return Structure.objects.none()

        if self.notif_type in {
            NotificationType.STRUCTURE_ONLINE,
            NotificationType.STRUCTURE_FUEL_ALERT,
            NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
            NotificationType.STRUCTURE_REFUELED_EXTRA,
            NotificationType.STRUCTURE_SERVICES_OFFLINE,
            NotificationType.STRUCTURE_WENT_LOW_POWER,
            NotificationType.STRUCTURE_WENT_HIGH_POWER,
            NotificationType.STRUCTURE_UNANCHORING,
            NotificationType.STRUCTURE_UNDER_ATTACK,
            NotificationType.STRUCTURE_LOST_SHIELD,
            NotificationType.STRUCTURE_LOST_ARMOR,
            NotificationType.STRUCTURE_DESTROYED,
            NotificationType.OWNERSHIP_TRANSFERRED,
            NotificationType.STRUCTURE_ANCHORING,
            NotificationType.MOONMINING_EXTRACTION_STARTED,
            NotificationType.MOONMINING_EXTRACTION_FINISHED,
            NotificationType.MOONMINING_AUTOMATIC_FRACTURE,
            NotificationType.MOONMINING_EXTRACTION_CANCELLED,
            NotificationType.MOONMINING_LASER_FIRED,
        }:
            try:
                structure_id = self._parsed_text["structureID"]
            except KeyError:
                qs = Structure.objects.none()
            else:
                qs = Structure.objects.filter(id=structure_id)

        elif self.notif_type == NotificationType.STRUCTURE_REINFORCEMENT_CHANGED:
            try:
                structure_ids = [
                    structure_info[0]
                    for structure_info in self._parsed_text["allStructureInfo"]
                ]
            except KeyError:
                qs = Structure.objects.none()
            else:
                qs = Structure.objects.filter(id__in=structure_ids)

        elif self.notif_type in {
            NotificationType.ORBITAL_ATTACKED,
            NotificationType.ORBITAL_REINFORCED,
        }:
            try:
                qs = Structure.objects.filter(
                    eve_planet_id=self._parsed_text["planetID"],
                    eve_type_id=self._parsed_text["typeID"],
                )
            except KeyError:
                qs = Structure.objects.none()

        elif self.notif_type in {
            NotificationType.TOWER_ALERT_MSG,
            NotificationType.TOWER_RESOURCE_ALERT_MSG,
            NotificationType.TOWER_REFUELED_EXTRA,
        }:
            try:
                qs = Structure.objects.filter(
                    eve_moon_id=self._parsed_text["moonID"],
                    eve_type_id=self._parsed_text["typeID"],
                )
            except KeyError:
                qs = Structure.objects.none()

        else:
            qs = Structure.objects.none()

        return qs

    def send_to_webhook(self, webhook: Webhook) -> bool:
        """Send this notification to a webhook.

        Return True if successful, else False.
        """
        logger.info("%s: Trying to sent to webhook: %s", self, webhook)
        try:
            embed, ping_type = self._generate_embed(webhook.language_code)
        except OSError as ex:
            logger.error("%s: Failed to generate embed: %s", self, ex, exc_info=True)
            return False

        content = self._create_content_with_pings(webhook, ping_type)
        content += self._add_discord_group_pings(webhook)
        max_content = dhooks_lite.Webhook.MAX_CHARACTERS
        if content and len(content) > max_content:
            logger.error("%s: Content of notification is too long: %s", self, content)
            return False

        username, avatar_url = self._gen_avatar()
        new_queue_size = webhook.send_message(
            content=content, embeds=[embed], username=username, avatar_url=avatar_url
        )
        success = new_queue_size > 0
        if success and not self.is_temporary:
            self.is_sent = True
            self.save()
        return success

    def _create_content_with_pings(self, webhook, ping_type):
        if webhook.has_default_pings_enabled and self.owner.has_default_pings_enabled:
            if ping_type == Webhook.PingType.EVERYONE:
                return "@everyone"

            if ping_type == Webhook.PingType.HERE:
                return "@here"

        return ""

    def _add_discord_group_pings(self, webhook) -> str:
        if not webhook.ping_groups.exists() and not self.owner.ping_groups.exists():
            return ""

        if "discord" not in app_labels():
            return ""

        DiscordUser = self._import_discord()
        groups = set(self.owner.ping_groups.all()) | set(webhook.ping_groups.all())
        content = ""
        for group in groups:
            try:
                role = DiscordUser.objects.group_to_role(group)
            except HTTPError:
                logger.warning("Failed to get Discord roles", exc_info=True)
            else:
                if role:
                    content += f" <@&{role['id']}>"

        return content

    def _generate_embed(
        self, language_code: Optional[str]
    ) -> Tuple[dhooks_lite.Embed, Optional[Webhook.PingType]]:
        """Generates a Discord embed for this notification."""
        from structures.core.notification_embeds import NotificationBaseEmbed

        logger.info("Creating embed with language = %s", language_code)
        with translation.override(language_code):
            notification_embed = NotificationBaseEmbed.create(self)
            embed = notification_embed.generate_embed()
            return embed, notification_embed.ping_type

    @staticmethod
    def _import_discord() -> object:
        from allianceauth.services.modules.discord.models import DiscordUser

        return DiscordUser

    def _gen_avatar(self) -> Tuple[Optional[str], Optional[str]]:
        if STRUCTURES_NOTIFICATION_SET_AVATAR:
            username = "Notifications"
            avatar_url = static_file_absolute_url("structures/img/structures_logo.png")
        else:
            username = None
            avatar_url = None
        avatar_url = avatar_url if is_absolute_url(avatar_url) else None
        return username, avatar_url

    def add_or_remove_timer(self) -> bool:
        """Add/remove a timer related to this notification for some types.

        Returns True when timers where added or removed, else False
        """
        if (
            not STRUCTURES_ADD_TIMERS
            or self.notif_type not in NotificationType.relevant_for_timerboard()
        ):
            return False

        from structures.core import notification_timers

        try:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                return notification_timers.add_or_remove_timer(self)
        except OSError as ex:
            logger.warning(
                "%s: Failed to add timer from notification: %s",
                self,
                ex,
                exc_info=True,
            )
        return False


class Notification(NotificationBase):
    """A notification in Eve Online.

    Notifications are usually created from Eve Online notifications received from ESI,
    but they can also be generated directly by Structures.
    """

    TEMPORARY_NOTIFICATION_ID = 999999999999

    notification_id = models.PositiveBigIntegerField(verbose_name=_("id"))
    created = models.DateTimeField(
        null=True,
        default=None,
        verbose_name=_("created"),
        help_text=_("Date when this notification was first received from ESI"),
    )
    is_read = models.BooleanField(
        null=True,
        default=None,
        verbose_name=_("is read"),
        help_text=_("True when this notification has read in the eve client"),
    )
    last_updated = models.DateTimeField(
        verbose_name=_("last updated"),
        help_text=_("Date when this notification has last been updated from ESI"),
    )
    sender = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_NULL,
        null=True,
        related_name="+",
        verbose_name=_("sender"),
    )

    text = models.TextField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("text"),
        help_text=_("Notification details in YAML"),
    )
    timestamp = models.DateTimeField(db_index=True, verbose_name=_("timestamp"))

    objects = NotificationManager()

    class Meta:
        verbose_name = _("eve notification")
        verbose_name_plural = _("eve notifications")
        unique_together = (("notification_id", "owner"),)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._parsed_text = yaml.safe_load(self.text) if self.text else {}

    def save(self, *args, **kwargs) -> None:
        if self.is_temporary:
            raise ValueError("Temporary notifications can not be saved")
        super().save(*args, **kwargs)

    @property
    def is_temporary(self) -> bool:
        return self.notification_id == self.TEMPORARY_NOTIFICATION_ID

    @property
    def is_generated(self) -> bool:
        return self.is_temporary

    # @classmethod
    # def get_all_types(cls) -> Set[int]:
    #     """returns a set with all supported notification types"""
    #     return {x[0] for x in NotificationType.choices}

    def parsed_text(self) -> dict:
        """Returns the notifications's text as dict."""
        return self._parsed_text

    def is_npc_attacking(self) -> bool:
        """Whether this notification is about a NPC attacking."""
        if self.notif_type in [
            NotificationType.ORBITAL_ATTACKED,
            NotificationType.STRUCTURE_UNDER_ATTACK,
        ]:
            parsed_text = self.parsed_text()
            corporation_id = None
            if self.notif_type == NotificationType.STRUCTURE_UNDER_ATTACK:
                if (
                    "corpLinkData" in parsed_text
                    and len(parsed_text["corpLinkData"]) >= 3
                ):
                    corporation_id = int(parsed_text["corpLinkData"][2])
            if self.notif_type == NotificationType.ORBITAL_ATTACKED:
                if "aggressorCorpID" in parsed_text:
                    corporation_id = int(parsed_text["aggressorCorpID"])
            if corporation_id:
                corporation = EveEntity(
                    category=EveEntity.CATEGORY_CORPORATION, id=corporation_id
                )
                return corporation.is_npc and not corporation.is_npc_starter_corporation
        return False

    def update_related_structures(self) -> bool:
        """Update related structure for this notification.

        Returns True if structures where updated, else False.
        """
        structures_qs = self.calc_related_structures()
        self.structures.clear()
        if structures_qs.exists():
            self.structures.add(*structures_qs)
            return True
        return False

    @classmethod
    def create_from_structure(
        cls, structure: Structure, notif_type: NotificationType, **kwargs
    ) -> "Notification":
        """Create new notification from given structure."""
        if "timestamp" not in kwargs:
            kwargs["timestamp"] = now()
        if "last_updated" not in kwargs:
            kwargs["last_updated"] = now()
        threshold = kwargs.pop("threshold") if "threshold" in kwargs else None
        if "text" not in kwargs:
            if notif_type in {
                NotificationType.STRUCTURE_FUEL_ALERT,
                NotificationType.STRUCTURE_REFUELED_EXTRA,
                NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
            }:
                data = {
                    "solarsystemID": structure.eve_solar_system_id,
                    "structureID": structure.id,
                    "structureTypeID": structure.eve_type_id,
                    "threshold": threshold,
                }
            elif notif_type in {
                NotificationType.TOWER_RESOURCE_ALERT_MSG,
                NotificationType.TOWER_REFUELED_EXTRA,
            }:
                data = {
                    "moonID": structure.eve_moon_id,
                    "typeID": structure.eve_type_id,
                    "structureID": structure.id,
                }
            else:
                raise ValueError("text property not provided and can not be generated.")
            kwargs["text"] = yaml.dump(data)
        if "sender" not in kwargs:
            sender, _ = EveEntity.objects.get_or_create_esi(id=EveCorporationId.DED)
            kwargs["sender"] = sender
        kwargs["notification_id"] = cls.TEMPORARY_NOTIFICATION_ID
        kwargs["owner"] = structure.owner
        kwargs["notif_type"] = notif_type
        return cls(**kwargs)


class GeneratedNotification(NotificationBase):
    """A notification generated by the Structures app, not by Eve Online."""

    details = models.JSONField(default=dict, verbose_name=_("details"))
    last_updated = models.DateTimeField(auto_now=True, verbose_name=_("last_updated"))
    timestamp = models.DateTimeField(auto_now_add=True, verbose_name=_("timestamp"))

    objects = GeneratedNotificationManager()

    class Meta:
        verbose_name = _("generated notification")
        verbose_name_plural = _("generated  notifications")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._parsed_text = self.details

    @property
    def notification_id(self) -> int:
        """Return the notification ID."""
        return self.pk

    @property
    def is_temporary(self) -> bool:
        return False

    @property
    def is_generated(self) -> bool:
        return True

    def is_npc_attacking(self) -> bool:
        return False


class BaseFuelAlertConfig(models.Model):
    """Configuration of structure fuel notifications."""

    channel_ping_type = models.CharField(
        max_length=2,
        choices=Webhook.PingType.choices,
        default=Webhook.PingType.HERE,
        verbose_name=_("channel pings"),
        help_text=_(
            "Option to ping every member of the channel. "
            "This setting can be overruled by the respective owner "
            "or webhook configuration"
        ),
    )
    color = models.IntegerField(
        choices=Webhook.Color.choices,
        default=Webhook.Color.WARNING,
        blank=True,
        null=True,
        verbose_name=_("color"),
        help_text=_("Context color of these notification on Discord"),
    )
    is_enabled = models.BooleanField(
        default=True,
        verbose_name=_("is_enabled"),
        help_text=_("Disabled configurations will not create any new alerts."),
    )

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return f"#{self.pk}"

    def send_new_notifications(self, force: bool = False) -> None:
        """Send new fuel notifications based on this config."""
        raise NotImplementedError()

    @staticmethod
    def relevant_webhooks() -> models.QuerySet:
        """Webhooks relevant for processing fuel notifications based on this config."""
        raise NotImplementedError()


class FuelAlertConfig(BaseFuelAlertConfig):
    """Configuration of structure fuel notifications."""

    end = models.PositiveIntegerField(
        verbose_name=_("end"), help_text=_("End of alerts in hours before fuel expires")
    )
    repeat = models.PositiveIntegerField(
        verbose_name=_("repeat"),
        help_text=_(
            "Notifications will be repeated every x hours. Set to 0 for no repeats"
        ),
    )
    start = models.PositiveIntegerField(
        verbose_name=_("start"),
        help_text=_("Start of alerts in hours before fuel expires"),
    )

    class Meta:
        verbose_name = _("structure fuel alert config")
        verbose_name_plural = _("structure fuel alert configs")

    def clean(self) -> None:
        if self.start is None or self.end is None or self.repeat is None:  # Fixes #83
            return  # these will be caught by the form validation later
        if self.start <= self.end:
            raise ValidationError(
                _("Start must be before end, i.e. have a larger value.")
            )
        if self.repeat >= self.start - self.end:
            raise ValidationError(
                {"repeat": _("Repeat can not be larger that the interval size.")}
            )
        new = range(self.end, self.start)
        for config in FuelAlertConfig.objects.exclude(pk=self.pk):
            current = range(config.end, config.start)
            overlap = range(max(new[0], current[0]), min(new[-1], current[-1]) + 1)
            if len(overlap) > 0:
                raise ValidationError(
                    _(
                        "This configuration may not overlap with an "
                        "existing configuration."
                    )
                )

    def save(self, *args, **kwargs) -> None:
        try:
            old_instance = FuelAlertConfig.objects.get(pk=self.pk)
        except FuelAlertConfig.DoesNotExist:
            old_instance = None
        super().save(*args, **kwargs)
        if old_instance and (
            old_instance.start != self.start
            or old_instance.end != self.end
            or old_instance.repeat != self.repeat
        ):
            self.structure_fuel_alerts.all().delete()

    def send_new_notifications(self, force: bool = False) -> None:
        """Send new fuel notifications based on this config."""
        structures = Structure.objects.filter(
            eve_type__eve_group__eve_category_id__in=[
                EveCategoryId.STARBASE,
                EveCategoryId.STRUCTURE,
            ],
            fuel_expires_at__isnull=False,
        )
        for structure in structures:
            if not structure.is_burning_fuel:
                continue
            hours_left = structure.hours_fuel_expires
            if self.start >= hours_left >= self.end:
                hours_last_alert = (
                    self.start
                    - (
                        math.floor((self.start - hours_left) / self.repeat)
                        * self.repeat
                    )
                    if self.repeat
                    else self.start
                )
                notif, created = FuelAlert.objects.get_or_create(
                    structure=structure, config=self, hours=hours_last_alert
                )
                if created or force:
                    notif.send_generated_notification()

    @staticmethod
    def relevant_webhooks() -> models.QuerySet:
        """Webhooks relevant for processing fuel notifications based on this config."""
        return (
            Webhook.objects.filter(is_active=True)
            .filter(Q(owners__isnull=False) | Q(structures__isnull=False))
            .filter(
                Q(notification_types__contains=NotificationType.STRUCTURE_FUEL_ALERT)
                | Q(
                    notification_types__contains=NotificationType.TOWER_RESOURCE_ALERT_MSG
                )
            )
            .distinct()
        )


class JumpFuelAlertConfig(BaseFuelAlertConfig):
    """Configuration of jump fuel notifications."""

    threshold = models.PositiveIntegerField(
        verbose_name=_("threshold"),
        help_text=_(
            "Notifications will be sent once fuel level in units reaches this threshold"
        ),
    )

    class Meta:
        verbose_name = _("jump fuel alert config")
        verbose_name_plural = _("jump fuel alert configs")

    def save(self, *args, **kwargs) -> None:
        try:
            old_instance = JumpFuelAlertConfig.objects.get(pk=self.pk)
        except JumpFuelAlertConfig.DoesNotExist:
            old_instance = None
        super().save(*args, **kwargs)
        if old_instance and (old_instance.threshold != self.threshold):
            self.jump_fuel_alerts.all().delete()

    def send_new_notifications(self, force: bool = False) -> None:
        """Send new fuel notifications based on this config."""
        jump_gates = Structure.objects.filter(eve_type_id=EveTypeId.JUMP_GATE)
        for jump_gate in jump_gates:
            if not jump_gate.is_burning_fuel:
                continue
            fuel_quantity = jump_gate.jump_fuel_quantity()
            if fuel_quantity and fuel_quantity < self.threshold:
                notif, created = JumpFuelAlert.objects.get_or_create(
                    structure=jump_gate, config=self
                )
                if created or force:
                    notif.send_generated_notification()

    @staticmethod
    def relevant_webhooks() -> models.QuerySet:
        """Webhooks relevant for processing jump fuel notifications based on this config."""
        return Webhook.objects.filter(
            is_active=True,
            notification_types__contains=NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
        ).filter(Q(owners__isnull=False) | Q(structures__isnull=False))


class BaseFuelAlert(models.Model):
    """Base class for fuel alerts."""

    class Meta:
        abstract = True

    def send_generated_notification(self):
        """Send a generated notification."""
        raise NotImplementedError()


class FuelAlert(BaseFuelAlert):
    """A generated notification alerting about fuel getting low in structures."""

    structure = models.ForeignKey(
        Structure,
        on_delete=models.CASCADE,
        related_name="structure_fuel_alerts",
        verbose_name=_("structure"),
    )
    config = models.ForeignKey(
        FuelAlertConfig,
        on_delete=models.CASCADE,
        related_name="structure_fuel_alerts",
        verbose_name=_("configuration"),
    )
    hours = models.PositiveIntegerField(
        db_index=True,
        verbose_name=_("hours"),
        help_text=_("Number of hours before fuel expiration this alert was sent"),
    )

    class Meta:
        verbose_name = _("structure fuel alert")
        verbose_name_plural = _("structure fuel alerts")
        constraints = [
            models.UniqueConstraint(
                fields=["structure", "config", "hours"],
                name="functional_pk_fuelalert",
            )
        ]

    def __str__(self) -> str:
        return f"{self.structure}-{self.config}-{self.hours}"

    def send_generated_notification(self):
        notif_type = (
            NotificationType.TOWER_RESOURCE_ALERT_MSG
            if self.structure.is_starbase
            else NotificationType.STRUCTURE_FUEL_ALERT
        )
        notif = Notification.create_from_structure(
            structure=self.structure, notif_type=notif_type
        )
        notif.send_to_configured_webhooks(
            ping_type_override=self.config.channel_ping_type,
            use_color_override=True,
            color_override=self.config.color,
        )


class JumpFuelAlert(BaseFuelAlert):
    """A generated notification alerting about jump fuel getting low."""

    structure = models.ForeignKey(
        Structure,
        on_delete=models.CASCADE,
        related_name="jump_fuel_alerts",
        verbose_name=_("structure"),
    )
    config = models.ForeignKey(
        JumpFuelAlertConfig,
        on_delete=models.CASCADE,
        related_name="jump_fuel_alerts",
        verbose_name=_("configuration"),
    )

    class Meta:
        verbose_name = _("jump fuel alert")
        verbose_name_plural = _("jump fuel alerts")

    def __str__(self) -> str:
        return f"{self.structure}-{self.config}"

    def send_generated_notification(self) -> None:
        notif = Notification.create_from_structure(
            structure=self.structure,
            notif_type=NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
            threshold=self.config.threshold,
        )
        notif.send_to_configured_webhooks(
            ping_type_override=self.config.channel_ping_type,
            use_color_override=True,
            color_override=self.config.color,
        )



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now, utc

from app_utils.esi_testing import EsiClientStub, EsiEndpoint
from app_utils.testing import NoSocketsTestCase

from structures.constants import EveCorporationId
from structures.core.notification_types import NotificationType
from structures.models import Structure, StructureService
from structures.tests import to_json
from structures.tests.testdata.factories import (
    EveEntityCorporationFactory,
    FuelAlertConfigFactory,
    OwnerFactory,
    StructureFactory,
    StructureServiceFactory,
    StructureTagFactory,
    UserMainDefaultOwnerFactory,
    WebhookFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.owners"


@patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
@patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
@patch(MODULE_PATH + ".esi")
class TestUpdateStructuresEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, structures_last_update_at=None)
        cls.corporation_id = cls.owner.corporation.corporation_id
        EveEntityCorporationFactory(
            id=EveCorporationId.DED, name="DED"
        )  # for notifications
        cls.endpoints = [
            EsiEndpoint(
                "Assets",
                "get_corporations_corporation_id_assets",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {
                            "is_singleton": False,
                            "item_id": 1300000001001,
                            "location_flag": "QuantumCoreRoom",
                            "location_id": 1000000000001,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 56201,  # Astrahus Upwell Quantum Core
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1300000001002,
                            "location_flag": "ServiceSlot0",
                            "location_id": 1000000000001,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 35894,  # Standup Cloning Center I
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1300000002001,
                            "location_flag": "ServiceSlot0",
                            "location_id": 1000000000002,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 35894,  # Standup Cloning Center I
                        },
                    ],
                    "2102": [
                        {
                            "is_singleton": False,
                            "item_id": 1300000003001,
                            "location_flag": "StructureFuel",
                            "location_id": 1000000000004,
                            "location_type": "item",
                            "quantity": 5000,
                            "type_id": 16273,  # Liquid Ozone
                        }
                    ],
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_structures",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {
                            "corporation_id": cls.corporation_id,
                            "fuel_expires": dt.datetime(2020, 3, 5, 5, tzinfo=utc),
                            "next_reinforce_apply": None,
                            "next_reinforce_hour": None,
                            "profile_id": 52436,
                            "reinforce_hour": 19,
                            "services": [
                                {"name": "Reprocessing", "state": "online"},
                                {"name": "Moon Drilling", "state": "online"},
                            ],
                            "state": "shield_vulnerable",
                            "state_timer_end": None,
                            "state_timer_start": None,
                            "structure_id": 1000000000002,
                            "system_id": 30002537,
                            "type_id": 35835,  # Athanor
                            "unanchors_at": None,
                        },
                        {
                            "corporation_id": cls.corporation_id,
                            "fuel_expires": dt.datetime(2020, 3, 5, 5, tzinfo=utc),
                            "next_reinforce_apply": None,
                            "next_reinforce_hour": None,
                            "profile_id": 101853,
                            "reinforce_hour": 18,
                            "services": [
                                {"name": "Clone Bay", "state": "online"},
                                {"name": "Market Hub", "state": "offline"},
                            ],
                            "state": "shield_vulnerable",
                            "state_timer_end": dt.datetime(2020, 4, 5, 7, tzinfo=utc),
                            "state_timer_start": dt.datetime(
                                2020, 4, 5, 6, 30, tzinfo=utc
                            ),
                            "structure_id": 1000000000001,
                            "system_id": 30002537,
                            "type_id": 35832,  # Astrahus
                            "unanchors_at": dt.datetime(2020, 5, 5, 6, 30, tzinfo=utc),
                        },
                        {
                            "corporation_id": cls.corporation_id,
                            "fuel_expires": None,
                            "next_reinforce_apply": None,
                            "next_reinforce_hour": None,
                            "profile_id": 101853,
                            "reinforce_hour": 18,
                            "services": None,
                            "state": "shield_vulnerable",
                            "state_timer_end": None,
                            "state_timer_start": None,
                            "structure_id": 1000000000003,
                            "system_id": 30000476,
                            "type_id": 35832,  # Astrahus
                            "unanchors_at": None,
                        },
                    ],
                },
            ),
            EsiEndpoint(
                "Universe",
                "get_universe_structures_structure_id",
                "structure_id",
                needs_token=True,
                data={
                    "1000000000001": {
                        "corporation_id": cls.corporation_id,
                        "name": "Amamake - Test Structure Alpha",
                        "position": {
                            "x": 55028384780.0,
                            "y": 7310316270.0,
                            "z": -163686684205.0,
                        },
                        "solar_system_id": 30002537,
                        "type_id": 35832,  # Astrahus
                    },
                    "1000000000002": {
                        "corporation_id": cls.corporation_id,
                        "name": "Amamake - Test Structure Bravo",
                        "position": {
                            "x": -2518743930339.066,
                            "y": -130157937025.56424,
                            "z": -442026427345.6355,
                        },
                        "solar_system_id": 30002537,
                        "type_id": 35835,  # Athanor
                    },
                    "1000000000003": {
                        "corporation_id": cls.corporation_id,
                        "name": "Amamake - Test Structure Charlie",
                        "position": {
                            "x": -2518743930339.066,
                            "y": -130157937025.56424,
                            "z": -442026427345.6355,
                        },
                        "solar_system_id": 30000476,
                        "type_id": 35832,  # Astrahus
                    },
                },
            ),
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(cls.endpoints)

    def test_can_sync_upwell_structures(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)
        self.assertAlmostEqual(
            owner.structures_last_update_at, now(), delta=dt.timedelta(seconds=30)
        )

        # must contain all expected structures
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

        # verify attributes for structure
        structure = Structure.objects.get(id=1000000000001)
        self.assertEqual(structure.name, "Test Structure Alpha")
        self.assertEqual(structure.position_x, 55028384780.0)
        self.assertEqual(structure.position_y, 7310316270.0)
        self.assertEqual(structure.position_z, -163686684205.0)
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(structure.eve_type_id, 35832)
        self.assertEqual(
            int(structure.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(structure.state, Structure.State.SHIELD_VULNERABLE)
        self.assertEqual(structure.reinforce_hour, 18)
        self.assertEqual(
            structure.fuel_expires_at, dt.datetime(2020, 3, 5, 5, 0, 0, tzinfo=utc)
        )
        self.assertEqual(
            structure.state_timer_start, dt.datetime(2020, 4, 5, 6, 30, 0, tzinfo=utc)
        )
        self.assertEqual(
            structure.state_timer_end, dt.datetime(2020, 4, 5, 7, 0, 0, tzinfo=utc)
        )
        self.assertEqual(
            structure.unanchors_at, dt.datetime(2020, 5, 5, 6, 30, 0, tzinfo=utc)
        )

        # must have created services with localizations
        # structure 1000000000001
        expected = {
            to_json(
                {
                    "name": "Clone Bay",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Clone Bay_zh",
                    "state": StructureService.State.ONLINE,
                }
            ),
            to_json(
                {
                    "name": "Market Hub",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Market Hub_zh",
                    "state": StructureService.State.OFFLINE,
                }
            ),
        }
        structure = Structure.objects.get(id=1000000000001)
        services = {
            to_json(
                {
                    "name": x.name,
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": x.name_zh,
                    "state": x.state,
                }
            )
            for x in structure.services.all()
        }
        self.assertEqual(services, expected)

        # must have created services with localizations
        # structure 1000000000002
        expected = {
            to_json(
                {
                    "name": "Reprocessing",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Reprocessing_zh",
                    "state": StructureService.State.ONLINE,
                }
            ),
            to_json(
                {
                    "name": "Moon Drilling",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Moon Drilling_zh",
                    "state": StructureService.State.ONLINE,
                }
            ),
        }
        structure = Structure.objects.get(id=1000000000002)
        services = {
            to_json(
                {
                    "name": x.name,
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": x.name_zh,
                    "state": x.state,
                }
            )
            for x in structure.services.all()
        }
        self.assertEqual(services, expected)

    def test_can_handle_owner_without_structures(self, mock_esi):
        # given
        owner = OwnerFactory(structures_last_update_at=None)
        corporation_id = owner.corporation.corporation_id
        endpoints = [
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_structures",
                "corporation_id",
                needs_token=True,
                data={f"{corporation_id}": []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)

        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)
        self.assertSetEqual(owner.structures.ids(), set())

    def test_should_not_break_when_endpoint_for_fetching_upwell_structures_is_down(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_structures",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = set()
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_update_will_not_break_on_http_error_from_structure_info(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Universe", "get_universe_structures_structure_id", http_error_code=500
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)
        structure = Structure.objects.get(id=1000000000002)
        self.assertEqual(structure.name, "(no data)")

    def test_update_will_not_break_on_403_error_from_structure_info(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Universe", "get_universe_structures_structure_id", http_error_code=403
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)
        structure = Structure.objects.get(id=1000000000002)
        self.assertEqual(structure.name, "(no data)")

    @patch(MODULE_PATH + ".Structure.objects.update_or_create_from_dict")
    def test_update_will_not_break_on_http_error_when_creating_structures(
        self, mock_create_structure, mock_esi
    ):
        mock_create_structure.side_effect = OSError
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)

    def test_should_remove_old_upwell_structures(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000004, name="delete-me")
        # when
        owner.update_structures_esi()
        # then
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_tags_are_not_modified_by_update(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then

        # should contain the right structures
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

        # adding tags
        tag_a = StructureTagFactory(name="tag_a")
        s = Structure.objects.get(id=1000000000001)
        s.tags.add(tag_a)
        s.save()

        # run update task 2nd time
        owner.update_structures_esi()

        # should still contain alls structures
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

        # should still contain the tag
        s_new = Structure.objects.get(id=1000000000001)
        self.assertEqual(s_new.tags.get(name="tag_a"), tag_a)

    def test_should_not_delete_existing_upwell_structures_when_update_failed(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_structures",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000001)
        StructureFactory(owner=owner, id=1000000000002)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = {1000000000001, 1000000000002}
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_should_remove_outdated_services(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        structure = StructureFactory(owner=owner, id=1000000000002)
        StructureServiceFactory(structure=structure, name="Clone Bay")
        # when
        owner.update_structures_esi()
        # then
        structure.refresh_from_db()
        services = {
            obj.name for obj in StructureService.objects.filter(structure=structure)
        }
        self.assertEqual(services, {"Moon Drilling", "Reprocessing"})

    @patch(
        "structures.models.structures_1.STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True
    )
    @patch("structures.models.notifications.Webhook.send_message")
    def test_should_send_refueled_notification_when_fuel_level_increased(
        self, mock_send_message, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        owner.webhooks.add(webhook)
        owner.update_structures_esi()
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = dt.datetime(2020, 3, 3, 0, 0, tzinfo=utc)
        structure.save()
        # when
        with patch("structures.models.structures_1.now") as now:
            now.return_value = dt.datetime(2020, 3, 2, 0, 0, tzinfo=utc)
            owner.update_structures_esi()
        # then
        self.assertTrue(mock_send_message.called)

    @patch(
        "structures.models.structures_1.STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True
    )
    @patch("structures.models.notifications.Webhook.send_message")
    def test_should_not_send_refueled_notification_when_fuel_level_unchanged(
        self, mock_send_message, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_send_message.side_effect = RuntimeError
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        owner.webhooks.add(webhook)
        with patch("structures.models.structures_1.now") as now:
            now.return_value = dt.datetime(2020, 3, 2, 0, 0, tzinfo=utc)
            owner.update_structures_esi()
            # when
            owner.update_structures_esi()
        # then
        self.assertFalse(mock_send_message.called)

    @patch("structures.models.notifications.Webhook.send_message")
    def test_should_remove_outdated_fuel_alerts_when_fuel_level_changed(
        self, mock_send_message, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        owner.webhooks.add(webhook)
        owner.update_structures_esi()
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = dt.datetime(2020, 3, 3, 0, 0, tzinfo=utc)
        structure.save()
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure.structure_fuel_alerts.create(config=config, hours=12)
        # when
        with patch("structures.models.structures_1.now") as now:
            now.return_value = dt.datetime(2020, 3, 2, 0, 0, tzinfo=utc)
            owner.update_structures_esi()
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)



"""Structure related models for Structures."""

import datetime as dt
import math
import re
from typing import List, Optional

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models, transaction
from django.db.models import Min, Sum
from django.utils.functional import cached_property
from django.utils.html import escape
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.views import bootstrap_label_html

from structures import __title__
from structures.app_settings import STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS
from structures.constants import EveCategoryId, EveGroupId, EveTypeId
from structures.core import starbases
from structures.helpers import datetime_almost_equal, hours_until_deadline
from structures.managers import StructureManager, StructureTagManager

from .eveuniverse import EveSpaceType

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class StructureTag(models.Model):
    """Tag for organizing structures."""

    # special tags
    NAME_SOV_TAG = gettext_noop("sov")
    NAME_HIGHSEC_TAG = gettext_noop("highsec")
    NAME_LOWSEC_TAG = gettext_noop("lowsec")
    NAME_NULLSEC_TAG = gettext_noop("nullsec")
    NAME_W_SPACE_TAG = gettext_noop("w-space")

    class Style(models.TextChoices):
        """A boostrap like style."""

        GREY = "default", _("grey")
        DARK_BLUE = "primary", _("dark blue")
        GREEN = "success", _("green")
        LIGHT_BLUE = "info", _("light blue")
        ORANGE = "warning", _("orange")
        RED = "danger", _("red")

    SPACE_TYPE_MAP = {
        EveSpaceType.HIGHSEC: {"name": NAME_HIGHSEC_TAG, "style": Style.GREEN},
        EveSpaceType.LOWSEC: {"name": NAME_LOWSEC_TAG, "style": Style.ORANGE},
        EveSpaceType.NULLSEC: {"name": NAME_NULLSEC_TAG, "style": Style.RED},
        EveSpaceType.W_SPACE: {"name": NAME_W_SPACE_TAG, "style": Style.LIGHT_BLUE},
    }

    name = models.CharField(
        max_length=255,
        unique=True,
        verbose_name=_("name"),
        help_text=_("Name of the tag, which must be unique"),
    )
    description = models.TextField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("description"),
        help_text=_("Description for this tag"),
    )
    style = models.CharField(
        max_length=16,
        choices=Style.choices,
        default="default",
        blank=True,
        verbose_name=_("style"),
        help_text=_("Color style of tag"),
    )
    order = models.PositiveIntegerField(
        default=100,
        blank=True,
        validators=[MinValueValidator(100)],
        verbose_name=_("order"),
        help_text=_(
            "Number defining the order tags are shown. "
            "custom tags can not have an order below 100"
        ),
    )
    is_default = models.BooleanField(
        default=False,
        verbose_name=_("is default"),
        help_text=_(
            "When enabled this custom tag will automatically be added to new structures"
        ),
    )
    is_user_managed = models.BooleanField(
        default=True,
        verbose_name=_("is user managed"),
        help_text=_(
            "When disabled this tag is created and managed by the system "
            "and can not be modified by users"
        ),
    )

    objects = StructureTagManager()

    class Meta:
        verbose_name = _("structure tag")
        verbose_name_plural = _("structure tags")
        ordering = ["order", "name"]

    def __str__(self) -> str:
        return self.name

    def __repr__(self):
        return f"{self.__class__.__name__}(name='{self.name}')"

    @property
    def html(self) -> str:
        """Return HTML for this tag."""
        if self.is_user_managed:
            name = escape(self.name)
        else:
            name = _(self.name)
        return bootstrap_label_html(name, self.style)

    @classmethod
    def sorted(cls, tags: list, reverse: bool = False) -> list:
        """returns a sorted copy of the provided list of tags"""
        return sorted(tags, key=lambda x: x.name.lower(), reverse=reverse)


class Structure(models.Model):  # pylint: disable = too-many-public-methods
    """A structure in Eve Online."""

    FUEL_DATES_EQUAL_THRESHOLD_UPWELL = 1800
    """Threshold in seconds when two fuel expiry dates will be judged as different."""

    FUEL_DATES_EQUAL_THRESHOLD_STARBASE = 7200
    """high fluctuation due to estimating."""

    class State(models.IntegerChoices):
        """A state of a structure."""

        # states Upwell structures
        ANCHOR_VULNERABLE = 1, _("anchor vulnerable")
        ANCHORING = 2, _("anchoring")
        ARMOR_REINFORCE = 3, _("armor reinforce")
        ARMOR_VULNERABLE = 4, _("armor vulnerable")
        DEPLOY_VULNERABLE = 5, _("deploy vulnerable")
        FITTING_INVULNERABLE = 6, _("fitting invulnerable")
        HULL_REINFORCE = 7, _("hull reinforce")
        HULL_VULNERABLE = 8, _("hull vulnerable")
        ONLINE_DEPRECATED = 9, _("online deprecated")
        ONLINING_VULNERABLE = 10, _("onlining vulnerable")
        SHIELD_VULNERABLE = 11, _("shield vulnerable")
        UNANCHORED = 12, _("unanchored")
        # starbases
        POS_OFFLINE = 21, _("offline")
        POS_ONLINE = 22, _("online")
        POS_ONLINING = 23, _("onlining")
        POS_REINFORCED = 24, _("reinforced")
        POS_UNANCHORING = 25, _("unanchoring ")
        # other
        NA = 0, _("N/A")
        UNKNOWN = 13, _("unknown")

        @classmethod
        def from_esi_name(cls, esi_state_name: str) -> "Structure.State":
            """returns matching state for esi state name of Upwell structures"""
            states_esi_map = {
                "anchor_vulnerable": cls.ANCHOR_VULNERABLE,
                "anchoring": cls.ANCHORING,
                "armor_reinforce": cls.ARMOR_REINFORCE,
                "armor_vulnerable": cls.ARMOR_VULNERABLE,
                "deploy_vulnerable": cls.DEPLOY_VULNERABLE,
                "fitting_invulnerable": cls.FITTING_INVULNERABLE,
                "hull_reinforce": cls.HULL_REINFORCE,
                "hull_vulnerable": cls.HULL_VULNERABLE,
                "online_deprecated": cls.ONLINE_DEPRECATED,
                "onlining_vulnerable": cls.ONLINING_VULNERABLE,
                "shield_vulnerable": cls.SHIELD_VULNERABLE,
                "unanchored": cls.UNANCHORED,
                "offline": cls.POS_OFFLINE,
                "online": cls.POS_ONLINE,
                "onlining": cls.POS_ONLINING,
                "reinforced": cls.POS_REINFORCED,
                "unanchoring ": cls.POS_UNANCHORING,
            }
            return (
                states_esi_map[esi_state_name]
                if esi_state_name in states_esi_map
                else cls.UNKNOWN
            )

    class PowerMode(models.TextChoices):
        """A power mode of a structure."""

        FULL_POWER = "FU", _("Full Power")
        LOW_POWER = "LO", _("Low Power")
        ABANDONED = "AB", _("Abandoned")
        LOW_ABANDONED = "LA", _("Abandoned?")
        UNKNOWN = "UN", _("Unknown")

    id = models.BigIntegerField(
        primary_key=True,
        verbose_name=_("id"),
        help_text=_("The Item ID of the structure"),
    )

    created_at = models.DateTimeField(
        default=now,
        verbose_name=_("created at"),
        help_text=_("Date this structure was received from ESI for the first time"),
    )
    eve_moon = models.ForeignKey(
        EveMoon,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        blank=True,
        related_name="+",
        verbose_name=_("moon"),
        help_text=_("Moon next to this structure - if any"),
    )
    eve_planet = models.ForeignKey(
        EvePlanet,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        blank=True,
        related_name="+",
        verbose_name=_("planet"),
        help_text=_("Planet next to this structure - if any"),
    )
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("solar system"),
        help_text=_("Solar System the structure is located"),
    )
    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("type"),
        help_text=_("Type of the structure"),
    )
    fuel_expires_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("fuel expires at"),
        help_text=_("Date on which the structure will run out of fuel"),
    )
    has_fitting = models.BooleanField(
        null=True,
        default=None,
        blank=True,
        db_index=True,
        verbose_name=_("has fitting"),
        help_text="Whether the structure has a fitting",
    )
    has_core = models.BooleanField(
        null=True,
        default=None,
        blank=True,
        db_index=True,
        verbose_name=_("has core"),
        help_text="Whether the structure has a quantum core",
    )
    last_online_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("last online at"),
        help_text=_("Date this structure had any of it's services online"),
    )
    last_updated_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("last updated at"),
        help_text=_("Date this structure was last updated from the EVE server"),
    )
    name = models.CharField(
        max_length=255,
        verbose_name=_("name"),
        help_text=_("The full name of the structure"),
    )
    next_reinforce_hour = models.PositiveIntegerField(
        null=True,
        default=None,
        blank=True,
        validators=[MaxValueValidator(23)],
        verbose_name=_("next reinforce hour"),
        help_text=_(
            "The requested change to reinforce_hour that will take "
            "effect at the time shown by next_reinforce_apply"
        ),
    )
    next_reinforce_apply = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("next reinforce apply"),
        help_text=_(
            "The requested change to reinforce_hour that will take "
            "effect at the time shown by next_reinforce_apply"
        ),
    )
    owner = models.ForeignKey(
        "Owner",
        on_delete=models.CASCADE,
        related_name="structures",
        verbose_name=_("owner"),
        help_text=_("Corporation that owns the structure"),
    )
    reinforce_hour = models.PositiveIntegerField(
        validators=[MaxValueValidator(23)],
        null=True,
        default=None,
        blank=True,
        verbose_name=_("reinforce hour"),
        help_text=_(
            "The average hour of day that determines the time +/- some hours "
            "when the structure will randomly exit its reinforcement periods "
            "and become vulnerable to attack against its armor and/or hull. "
        ),
    )
    position_x = models.FloatField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("position x"),
        help_text=_("X coordinate of position in the solar system"),
    )
    position_y = models.FloatField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("position y"),
        help_text=_("Y coordinate of position in the solar system"),
    )
    position_z = models.FloatField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("position z"),
        help_text=_("Z coordinate of position in the solar system"),
    )
    state = models.IntegerField(
        choices=State.choices,
        default=State.UNKNOWN,
        blank=True,
        verbose_name=_("state"),
        help_text=_("Current state of the structure"),
    )
    state_timer_end = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("state timer end"),
        help_text=_("Date at which the structure entered it's current state"),
    )
    state_timer_start = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("state timer start"),
        help_text=_("Date at which the structure will move to it's next state"),
    )
    tags = models.ManyToManyField(
        StructureTag,
        default=None,
        blank=True,
        related_name="structures",
        verbose_name=_("tags"),
        help_text=_("List of tags for this structure. "),
    )
    unanchors_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("unanchors at"),
        help_text=_("Date at which the structure will unanchor"),
    )
    webhooks = models.ManyToManyField(
        "Webhook",
        default=None,
        blank=True,
        related_name="structures",
        verbose_name=_("webhooks"),
        help_text=_(
            "Webhooks for sending notifications to. "
            "If any webhook is enabled, these will be used instead of the webhooks "
            "defined for the respective owner. "
            "If no webhook is enabled the owner's setting will be used. "
        ),
    )

    objects = StructureManager()

    class Meta:
        verbose_name = _("structure")
        verbose_name_plural = _("structures")

    def __str__(self) -> str:
        if self.is_upwell_structure:
            try:
                location_name = self.eve_solar_system.name
            except AttributeError:
                location_name = "?"
        else:
            location_name = self.location_name
        return f"{location_name} - {self.name}" if self.name else str(location_name)

    def __repr__(self) -> str:
        try:
            eve_solar_system_name = self.eve_solar_system.name
        except AttributeError:
            eve_solar_system_name = ""
        try:
            eve_planet_name = self.eve_planet.name
        except AttributeError:
            eve_planet_name = ""
        try:
            eve_moon_name = self.eve_moon.name
        except AttributeError:
            eve_moon_name = ""
        return (
            f"{self.__class__.__name__}(id={self.id}, "
            f"eve_solar_system='{eve_solar_system_name}', "
            f"eve_planet='{eve_planet_name}', "
            f"eve_moon='{eve_moon_name}', "
            f"name='{self.name}')"
        )

    def save(self, *args, **kwargs):
        is_new = self._state.adding is True
        super().save(*args, **kwargs)
        if is_new:
            for tag in StructureTag.objects.filter(is_default=True):
                self.tags.add(tag)
        # make sure related objects are saved whenever structure is saved
        self.update_generated_tags()

    @property
    def is_jump_gate(self) -> bool:
        """Return True if this structure is a jump gate, else False."""
        return self.eve_type_id == EveTypeId.JUMP_GATE

    @property
    def is_poco(self) -> bool:
        """Return True if this structure is a customs office, else False."""
        return self.eve_type_id == EveTypeId.CUSTOMS_OFFICE

    @cached_property
    def is_starbase(self) -> bool:
        """Return True if this structure is a starbase, else False."""
        return starbases.is_starbase(self.eve_type)

    @property
    def is_skyhook(self) -> bool:
        """Return True if this structure is a skyhook, else False."""
        return self.eve_type_id == EveTypeId.ORBITAL_SKYHOOK

    @cached_property
    def is_upwell_structure(self) -> bool:
        """Return True if this structure is an upwell structure, else False."""
        return self.eve_type.eve_group.eve_category_id == EveCategoryId.STRUCTURE

    @property
    def is_full_power(self) -> Optional[bool]:
        """Return True if structure is full power, False if not.

        Returns None if state can not be determined
        """
        power_mode = self.power_mode
        if not power_mode:
            return None

        return power_mode == self.PowerMode.FULL_POWER

    @property
    def is_low_power(self) -> Optional[bool]:
        """Return True if structure is low power, False if not.

        Returns None if state can not be determined
        """
        power_mode = self.power_mode
        if not power_mode:
            return None
        return power_mode == self.PowerMode.LOW_POWER

    @property
    def is_abandoned(self) -> Optional[bool]:
        """return True if structure is abandoned, False if not.

        Returns None if state can not be determined
        """
        power_mode = self.power_mode
        if not power_mode:
            return None
        return power_mode == self.PowerMode.ABANDONED

    @property
    def is_maybe_abandoned(self) -> Optional[bool]:
        """return True if structure is maybe abandoned, False if not.

        Returns None if state can not be determined
        """
        power_mode = self.power_mode
        if not power_mode:
            return None
        return power_mode == self.PowerMode.LOW_ABANDONED

    @property
    def power_mode(self) -> Optional["PowerMode"]:
        """returns the calculated power mode of this structure, e.g. low power
        returns None for non upwell structures
        """
        if not self.is_upwell_structure:
            return None

        if self.fuel_expires_at and self.fuel_expires_at > now():
            return self.PowerMode.FULL_POWER

        if self.last_online_at:
            if self.last_online_at >= now() - dt.timedelta(days=7):
                return self.PowerMode.LOW_POWER
            return self.PowerMode.ABANDONED

        if self.state in {self.State.ANCHORING, self.State.ANCHOR_VULNERABLE}:
            return self.PowerMode.LOW_POWER

        return self.PowerMode.LOW_ABANDONED

    @property
    def is_reinforced(self) -> bool:
        """Return True if this structure is reinforced, else False."""
        return self.state in {
            self.State.ARMOR_REINFORCE,
            self.State.HULL_REINFORCE,
            self.State.POS_REINFORCED,
        }

    @property
    def is_burning_fuel(self) -> bool:
        """True when this structure is currently burning fuel, else False."""
        if self.is_upwell_structure and self.is_full_power:
            return True
        if self.is_starbase and self.state in [
            self.State.POS_ONLINE,
            self.State.POS_REINFORCED,
            self.State.POS_UNANCHORING,
        ]:
            return True
        return False

    @property
    def has_position(self) -> bool:
        """Evaluate if this structure has a position."""
        return (
            self.position_x is not None
            and self.position_y is not None
            and self.position_z is not None
        )

    @cached_property
    def location_name(self) -> str:
        """Name of this structures's location."""
        try:
            return self.eve_moon.name
        except AttributeError:
            pass
        try:
            return self.eve_planet.name
        except AttributeError:
            pass
        try:
            return self.eve_solar_system.name
        except AttributeError:
            return "?"

    def owner_has_sov(self) -> bool:
        """Return True if the owner of this structure has sov in this solar system,
        else False.
        """
        return self.owner.has_sov(self.eve_solar_system)

    def distance_to_object(self, x: float, y: float, z: float) -> float:
        """Distance to object with given coordinates (within same solar system)."""
        if not self.has_position:
            raise ValueError(
                f"{self}: Can not calculate distance from a structure without a position"
            )
        return math.sqrt(
            pow(x - self.position_x, 2)  # type: ignore
            + pow(y - self.position_y, 2)  # type: ignore
            + pow(z - self.position_z, 2)  # type: ignore
        )

    @cached_property
    def structure_fuel_quantity(self) -> Optional[int]:
        """Current quantity of fuel blocks in units used as fuel."""
        return self.items.filter(
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            eve_type__eve_group_id=EveGroupId.FUEL_BLOCK,
        ).aggregate(Sum("quantity"))["quantity__sum"]

    def jump_fuel_quantity(self) -> Optional[int]:
        """Current quantity of liquid ozone in units used as fuel."""
        return self.items.filter(
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            eve_type_id=EveTypeId.LIQUID_OZONE,
        ).aggregate(Sum("quantity"))["quantity__sum"]

    def structure_fuel_usage(self) -> Optional[int]:
        """Needed fuel blocks per day."""
        if not self.fuel_expires_at:
            return None

        fuel_quantity = self.structure_fuel_quantity
        if not fuel_quantity:
            return None

        assets_last_updated_at = self.items.filter(
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            eve_type__eve_group_id=EveGroupId.FUEL_BLOCK,
        ).aggregate(Min("last_updated_at"))["last_updated_at__min"]
        if not assets_last_updated_at:
            return None

        try:
            return math.ceil(
                fuel_quantity
                / hours_until_deadline(self.fuel_expires_at, assets_last_updated_at)
                * 24
            )
        except ZeroDivisionError:
            return None

    @property
    def hours_fuel_expires(self) -> Optional[float]:
        """Hours until fuel expires."""
        if self.fuel_expires_at:
            return hours_until_deadline(self.fuel_expires_at)
        return None

    def is_fuel_expiry_date_different(self, other: "Structure") -> bool:
        """True when fuel expiry date from other structure is different.

        Will compare using threshold setting.
        """
        change_threshold = (
            self.FUEL_DATES_EQUAL_THRESHOLD_UPWELL
            if self.is_upwell_structure
            else self.FUEL_DATES_EQUAL_THRESHOLD_STARBASE
        )
        return not other.fuel_expires_at or not datetime_almost_equal(
            other.fuel_expires_at, self.fuel_expires_at, change_threshold
        )

    def handle_fuel_notifications(self, old_instance: "Structure"):
        """Remove fuel notifications if fuel levels have changed
        and sent refueled notifications if structure has been refueled.
        """
        if self.fuel_expires_at and old_instance and self.pk == old_instance.pk:
            logger_tag = f"{self}: Fuel notifications"
            if self.fuel_expires_at != old_instance.fuel_expires_at:
                if old_instance.fuel_expires_at:
                    delta = int(
                        abs(
                            (
                                self.fuel_expires_at - old_instance.fuel_expires_at
                            ).total_seconds()
                        )
                    )
                else:
                    delta = "-"
                old_date = (
                    old_instance.fuel_expires_at.isoformat()
                    if old_instance.fuel_expires_at
                    else None
                )
                current_date = self.fuel_expires_at.isoformat()
                logger.info(
                    "%s: Fuel expiry dates changed: old|current|delta: %s|%s|%s",
                    logger_tag,
                    old_date,
                    current_date,
                    delta,
                )
            if (
                self.is_burning_fuel
                and old_instance.is_burning_fuel
                and self.is_fuel_expiry_date_different(old_instance)
            ):
                logger.info(
                    "%s: Structure fuel level has changed. "
                    "Therefore removing current fuel notifications.",
                    logger_tag,
                )
                self.structure_fuel_alerts.all().delete()
                if STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS and (
                    not old_instance.fuel_expires_at
                    or old_instance.fuel_expires_at < self.fuel_expires_at
                ):
                    logger.info("%s: Structure has been refueled.", logger_tag)
                    self._send_refueled_notification()

    def _send_refueled_notification(self):
        """Send a refueled notifications for this structure."""
        from .notifications import Notification, NotificationType

        notif_type = (
            NotificationType.TOWER_REFUELED_EXTRA
            if self.is_starbase
            else NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        notif = Notification.create_from_structure(
            structure=self, notif_type=notif_type
        )
        notif.send_to_configured_webhooks()

    def reevaluate_jump_fuel_alerts(self):
        """Remove outdated fuel alerts based on potentially changed fuel levels."""
        jump_fuel_quantity = self.jump_fuel_quantity()
        if jump_fuel_quantity:
            self.jump_fuel_alerts.filter(
                config__threshold__lt=jump_fuel_quantity
            ).delete()

    def get_power_mode_display(self) -> str:
        """Return this structure's power mode as label for display.
        Or return an empty string for structures that have no power mode.
        """
        return self.PowerMode(self.power_mode).label if self.power_mode else ""

    def update_generated_tags(self, recreate_tags=False):
        """Update generated tags for this structure.

        recreate_tags: when set true all tags will be re-created,
        otherwise just re-added if they are missing
        """
        method_name = (
            "update_or_create_for_space_type"
            if recreate_tags
            else "get_or_create_for_space_type"
        )

        try:
            space_type_tag, _ = getattr(StructureTag.objects, method_name)(
                self.eve_solar_system
            )
        except ValueError:
            logger.warning(
                "%s: Can not generate tag for unknown space type of this solar system.",
                self.eve_solar_system,
            )
        else:
            self.tags.add(space_type_tag)

        if self.owner_has_sov():
            method_name = (
                "update_or_create_for_sov" if recreate_tags else "get_or_create_for_sov"
            )

            sov_tag, _ = getattr(StructureTag.objects, method_name)()
            self.tags.add(sov_tag)

    def update_items(self, structure_items: List["StructureItem"]):
        """Update items for this structure."""
        with transaction.atomic():
            self.items.all().delete()
            if structure_items:
                # remove items that have been transferred between structures
                item_ids = {item.id for item in structure_items}
                StructureItem.objects.filter(id__in=item_ids).delete()
                # create new items
                self.items.bulk_create(structure_items)

    @classmethod
    def extract_name_from_esi_response(cls, esi_name):
        """extracts the structure's name from the name in an ESI response"""
        matches = re.search(r"^\S+ - (.+)", esi_name)
        return matches.group(1) if matches else esi_name


class StructureItem(models.Model):
    """An item in a structure"""

    class LocationFlag(models.TextChoices):
        """Frequently used location flags.
        This are not all, so can not be used as choices.
        """

        AUTOFIT = "AutoFit"
        CARGO = "Cargo"
        CORP_DELIVERIES = "CorpDeliveries"
        FIGHTER_BAY = "FighterBay"
        OFFICE_FOLDER = "OfficeFolder"
        QUANTUM_CORE_ROOM = "QuantumCoreRoom"
        SECONDARY_STORAGE = "SecondaryStorage"
        STRUCTURE_FUEL = "StructureFuel"

    id = models.BigIntegerField(
        primary_key=True, verbose_name=_("id"), help_text=_("The Eve item ID")
    )
    structure = models.ForeignKey(
        Structure,
        on_delete=models.CASCADE,
        related_name="items",
        verbose_name=_("structure"),
        help_text=_("Structure this item is located in"),
    )

    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("type"),
        help_text="Type of the item",
    )
    is_singleton = models.BooleanField(verbose_name=_("is singleton"))
    last_updated_at = models.DateTimeField(
        auto_now=True, verbose_name=_("last updated at")
    )
    location_flag = models.CharField(max_length=255, verbose_name=_("location flag"))
    quantity = models.IntegerField(verbose_name=_("quantity"))

    class Meta:
        verbose_name = _("structure item")
        verbose_name_plural = _("structure items")

    def __str__(self) -> str:
        return str(self.eve_type.name)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"pk={self.pk}, structure=<{self.structure}>, eve_type=<{self.eve_type}>"
            ")"
        )

    @classmethod
    def from_esi_asset(cls, item: dict, structure: "Structure") -> "StructureItem":
        """Create new object from ESI asset item."""
        eve_type, _ = EveType.objects.get_or_create_esi(id=item["type_id"])
        return StructureItem(
            id=item["item_id"],
            structure=structure,
            eve_type=eve_type,
            is_singleton=item["is_singleton"],
            location_flag=item["location_flag"],
            quantity=item["quantity"],
        )



"""App for managing Eve Online structures with Alliance Auth."""

# pylint: disable = invalid-name
default_app_config = "structures.apps.StructuresConfig"

__version__ = "2.15.0"
__title__ = "Structures"



"""Views for structure list page."""

import functools
from collections import defaultdict
from enum import Enum, IntEnum
from typing import Dict, Sequence, Set, Union
from urllib.parse import urlencode

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import User
from django.db.models import Prefetch
from django.http import HttpRequest, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.templatetags.static import static
from django.urls import reverse
from django.utils import translation
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from esi.decorators import token_required
from esi.models import Token
from eveuniverse.core import eveimageserver
from eveuniverse.models import EveType, EveTypeDogmaAttribute

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import is_night_mode, notify_admins
from app_utils.logging import LoggerAddTag

from structures import __title__, tasks
from structures.app_settings import (
    STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED,
    STRUCTURES_DEFAULT_LANGUAGE,
    STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED,
    STRUCTURES_SHOW_JUMP_GATES,
)
from structures.constants import EveAttributeId, EveCategoryId, EveGroupId, EveTypeId
from structures.core.serializers import StructureListSerializer
from structures.forms import TagsFilterForm
from structures.models import (
    Owner,
    Structure,
    StructureItem,
    StructureService,
    StructureTag,
    Webhook,
)

from .common import add_common_context, add_common_data_export

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

QUERY_PARAM_TAGS = "tags"


class StructureSelection(str, Enum):
    """A pre-defined selection to filter structures data."""

    STRUCTURES = "structures"
    ORBITALS = "orbitals"
    STARBASES = "starbases"
    JUMP_GATES = "jump_gates"
    ALL = "all"


def _urlencode_tags(tags: Sequence[StructureTag]) -> str:
    params = {QUERY_PARAM_TAGS: ",".join([tag.name for tag in tags])}
    params_encoded = urlencode(params)
    return params_encoded


@login_required
@permission_required("structures.basic_access")
def index(request: HttpRequest):
    """Redirect from index view to main."""

    if (
        request.user.has_perm("structures.view_corporation_structures")
        or request.user.has_perm("structures.view_alliance_structures")
        or request.user.has_perm("structures.view_all_structures")
    ):
        url = reverse("structures:structure_list")
        if STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED:
            tags = StructureTag.objects.filter(is_default=True)
            params_encoded = _urlencode_tags(tags)
            url += f"?{params_encoded}"
    else:
        url = reverse("structures:public")

    return redirect(url)


@login_required
@permission_required("structures.basic_access")
def structure_list(request: HttpRequest):
    """Render structure list view."""
    tags = []
    if request.method == "POST":
        form = TagsFilterForm(data=request.POST)
        if form.is_valid():
            for name, activated in form.cleaned_data.items():
                if activated:
                    tags.append(get_object_or_404(StructureTag, name=name))

            url = reverse("structures:structure_list")
            if tags:
                params_encoded = _urlencode_tags(tags)
                url += f"?{params_encoded}"
            return redirect(url)
    else:
        tags_raw = request.GET.get(QUERY_PARAM_TAGS)
        if tags_raw:
            tags_parsed = tags_raw.split(",")
            tags = list(StructureTag.objects.filter(name__in=tags_parsed))
        form = TagsFilterForm(initial={tag.name: True for tag in tags})

    structures_count = _structures_query(
        request.user, StructureSelection.STRUCTURES, tags
    ).count()
    orbitals_count = _structures_query(
        request.user, StructureSelection.ORBITALS, tags
    ).count()
    starbases_count = _structures_query(
        request.user, StructureSelection.STARBASES, tags
    ).count()
    jump_gates_count = _structures_query(
        request.user, StructureSelection.JUMP_GATES, tags
    ).count()

    data_export = add_common_data_export(_construct_data_export(request, tags))

    context = {
        "active_tags": tags,
        "tags_filter_form": form,
        "tags_exist": StructureTag.objects.exists(),
        "show_jump_gates_tab": STRUCTURES_SHOW_JUMP_GATES,
        "structures_count": structures_count,
        "orbitals_count": orbitals_count,
        "starbases_count": starbases_count,
        "jump_gates_count": jump_gates_count,
        "data_export": data_export,
    }
    return render(request, "structures/structures.html", add_common_context(context))


def _construct_data_export(request, tags):
    structures_ajax_url = _construct_ajax_url(StructureSelection.STRUCTURES, tags)
    pocos_ajax_url = _construct_ajax_url(StructureSelection.ORBITALS, tags)
    starbases_ajax_url = _construct_ajax_url(StructureSelection.STARBASES, tags)
    jump_gates_ajax_url = _construct_ajax_url(StructureSelection.JUMP_GATES, tags)

    if is_night_mode(request):
        spinner_image_url = static("structures/img/bars-rotate-fade-white-36.svg")
    else:
        spinner_image_url = static("structures/img/bars-rotate-fade-black-36.svg")

    data_export = {
        "structures_ajax_url": structures_ajax_url,
        "pocos_ajax_url": pocos_ajax_url,
        "starbases_ajax_url": starbases_ajax_url,
        "jump_gates_ajax_url": jump_gates_ajax_url,
        "spinner_image_url": spinner_image_url,
        "filter_titles": {
            "alliance": _("Alliance"),
            "corporation": _("Corporation"),
            "constellation": _("Constellation"),
            "core": _("Core?"),
            "group": _("Group"),
            "power_mode": _("Power Mode"),
            "region": _("Region"),
            "reinforced": _("Reinforced?"),
            "state": _("State"),
            "solar_system": _("Solar System"),
        },
    }

    return data_export


def _construct_ajax_url(selection: StructureSelection, tags):
    ajax_url = reverse("structures:structure_list_data", args=[selection.value])
    if tags:
        params_encoded = _urlencode_tags(tags)
        ajax_url += f"?{params_encoded}"
    return ajax_url


@login_required
@permission_required("structures.basic_access")
def structure_list_data(request: HttpRequest, selection: str) -> JsonResponse:
    """Return structure list in JSON for AJAX call in structure_list view."""
    tag_names = _current_tags(request)
    structures_qs = _structures_query(request.user, selection, tag_names)

    serializer = StructureListSerializer(queryset=structures_qs, request=request)
    return JsonResponse({"data": serializer.to_list()})


def _structures_query(
    user: User, selection: Union[StructureSelection, str], tag_names: Set[str]
):
    """Return query for a variant and user and active tags."""
    structures_qs = (
        Structure.objects.visible_for_user(user)
        .select_related_defaults()
        .filter_tags(tag_names)
    )

    selection = StructureSelection(selection)
    if selection == StructureSelection.STRUCTURES:
        structures_qs = structures_qs.filter(
            eve_type__eve_group__eve_category_id=EveCategoryId.STRUCTURE
        )

    elif selection == StructureSelection.ORBITALS:
        structures_qs = structures_qs.filter(
            eve_type__eve_group__eve_category_id=EveCategoryId.ORBITAL
        ).annotate_has_poco_details()

    elif selection == StructureSelection.STARBASES:
        structures_qs = structures_qs.filter(
            eve_type__eve_group__eve_category_id=EveCategoryId.STARBASE
        ).annotate_has_starbase_detail()

    elif selection == StructureSelection.JUMP_GATES:
        structures_qs = structures_qs.filter(
            eve_type=EveTypeId.JUMP_GATE
        ).annotate_jump_fuel_quantity()

    elif selection == StructureSelection.ALL:
        pass

    return structures_qs


def _current_tags(request) -> Set[str]:
    """Return currently enabled tags."""
    tags_raw = request.GET.get(QUERY_PARAM_TAGS)
    tags = tags_raw.split(",") if tags_raw else []
    return set(tags)


class FakeEveType:
    """A faked eve type."""

    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.profile_url = ""

    def icon_url(self, size=64) -> str:
        """Return icon url for an EveType."""
        return eveimageserver.type_icon_url(self.id, size)


class FakeAsset:
    """Fake asset object for showing additional information in the asset list."""

    def __init__(self, name, quantity, eve_type_id):
        self.name = name
        self.quantity = quantity
        self.eve_type_id = eve_type_id
        self.eve_type = FakeEveType(eve_type_id, name)
        self.is_singleton = False


class Slot(IntEnum):
    """A slot type in a fitting."""

    HIGH = 14
    MEDIUM = 13
    LOW = 12
    RIG = 1137
    SERVICE = 2056

    def image_url(self, type_attributes: dict) -> str:
        """Return url to image file for this slot variant"""
        id_map = {
            self.HIGH: "h",
            self.MEDIUM: "m",
            self.LOW: "l",
            self.RIG: "r",
            self.SERVICE: "s",
        }
        try:
            slot_num = type_attributes[self.value]
            my_id = id_map[Slot(self.value)]
            return static(f"structures/img/panel/{slot_num}{my_id}.png")
        except KeyError:
            return ""


@login_required
@permission_required("structures.view_structure_fit")
def structure_details(request: HttpRequest, structure_id: int):
    """Render structure details view."""

    structure: Structure = get_object_or_404(
        Structure.objects.select_related(
            "owner",
            "owner__corporation",
            "owner__corporation__alliance",
            "eve_type",
            "eve_type__eve_group",
            "eve_solar_system",
            "eve_solar_system__eve_constellation",
            "eve_solar_system__eve_constellation__eve_region",
        ).prefetch_related(
            Prefetch(
                "services",
                queryset=StructureService.objects.order_by("name"),
                to_attr="services_ordered",
            )
        ),
        id=structure_id,
    )
    assets = structure.items.select_related("eve_type", "eve_type__eve_group")
    high_slots = _extract_slot_assets(assets, "HiSlot")
    med_slots = _extract_slot_assets(assets, "MedSlot")
    low_slots = _extract_slot_assets(assets, "LoSlot")
    rig_slots = _extract_slot_assets(assets, "RigSlot")
    service_slots = _extract_slot_assets(assets, "ServiceSlot")
    fighter_tubes = _extract_slot_assets(assets, "FighterTube")
    _patch_fighter_tube_quantities(fighter_tubes)

    assets_grouped = _init_assets_grouped(assets)

    if structure.is_upwell_structure:
        assets_grouped["fuel_usage"] = [
            FakeAsset(
                name=_("Fuel blocks per day (est.)"),
                quantity=structure.structure_fuel_usage(),
                eve_type_id=24756,
            )
        ]

    fuel_blocks_total = (
        functools.reduce(
            lambda x, y: x + y, [obj.quantity for obj in assets_grouped["fuel_bay"]]
        )
        if assets_grouped["fuel_bay"]
        else 0
    )
    ammo_total = (
        functools.reduce(
            lambda x, y: x + y, [obj.quantity for obj in assets_grouped["ammo_hold"]]
        )
        if assets_grouped["ammo_hold"]
        else 0
    )
    ammo_total += _calc_fighters_total(fighter_tubes, assets_grouped)

    services = structure.services_ordered

    context = {
        "fitting": assets,
        "slots": _generate_slot_image_urls(structure),
        "slot_assets": {
            "high_slots": high_slots,
            "med_slots": med_slots,
            "low_slots": low_slots,
            "rig_slots": rig_slots,
            "service_slots": service_slots,
            "fighter_tubes": fighter_tubes,
        },
        "assets_grouped": assets_grouped,
        "structure": structure,
        "modules_count": len(
            high_slots + med_slots + low_slots + rig_slots + service_slots
        ),
        "fuel_blocks_total": fuel_blocks_total,
        "ammo_total": ammo_total,
        "last_updated": structure.owner.assets_last_update_at,
        "services": services,
        "services_count": len(services),
    }
    return render(request, "structures/modals/structure_details.html", context)


def _init_assets_grouped(assets):
    assets_grouped = {"ammo_hold": [], "fighter_bay": [], "fuel_bay": []}
    for asset in assets:
        if asset.location_flag == StructureItem.LocationFlag.CARGO:
            assets_grouped["ammo_hold"].append(asset)
        elif asset.location_flag == StructureItem.LocationFlag.FIGHTER_BAY:
            assets_grouped["fighter_bay"].append(asset)
        elif asset.location_flag == StructureItem.LocationFlag.STRUCTURE_FUEL:
            assets_grouped["fuel_bay"].append(asset)
        else:
            assets_grouped[asset.location_flag] = asset
    return assets_grouped


def _calc_fighters_total(fighter_tubes, assets_grouped):
    fighters_consolidated = assets_grouped["fighter_bay"] + fighter_tubes
    fighters_total = (
        functools.reduce(
            lambda x, y: x + y, [obj.quantity for obj in fighters_consolidated]
        )
        if fighters_consolidated
        else 0
    )

    return fighters_total


def _generate_slot_image_urls(structure):
    type_attributes = {
        obj["eve_dogma_attribute_id"]: int(obj["value"])
        for obj in EveTypeDogmaAttribute.objects.filter(
            eve_type_id=structure.eve_type_id
        ).values("eve_dogma_attribute_id", "value")
    }
    slot_image_urls = {
        "high": Slot.HIGH.image_url(type_attributes),
        "med": Slot.MEDIUM.image_url(type_attributes),
        "low": Slot.LOW.image_url(type_attributes),
        "rig": Slot.RIG.image_url(type_attributes),
        "service": Slot.SERVICE.image_url(type_attributes),
    }

    return slot_image_urls


def _extract_slot_assets(fittings: list, slot_name: str) -> list:
    """Return assets for slot sorted by slot number"""
    return [
        asset[0]
        for asset in sorted(
            [
                (asset, asset.location_flag[-1])
                for asset in fittings
                if asset.location_flag.startswith(slot_name)
            ],
            key=lambda x: x[1],
        )
    ]


def _patch_fighter_tube_quantities(fighter_tubes):
    eve_type_ids = {item.eve_type_id for item in fighter_tubes}
    eve_types = [
        EveType.objects.get_or_create_esi(
            id=eve_type_id, enabled_sections=[EveType.Section.DOGMAS]
        )[0]
        for eve_type_id in eve_type_ids
    ]
    squadron_sizes = {
        eve_type.id: int(
            eve_type.dogma_attributes.get(
                eve_dogma_attribute=EveAttributeId.SQUADRON_SIZE.value
            ).value
        )
        for eve_type in eve_types
    }
    for item in fighter_tubes:
        try:
            squadron_size = squadron_sizes[item.eve_type_id]
        except KeyError:
            pass
        else:
            item.quantity = squadron_size
            item.is_singleton = False


@login_required
@permission_required("structures.basic_access")
def poco_details(request: HttpRequest, structure_id):
    """Shows details modal for a POCO."""

    structure = get_object_or_404(
        Structure.objects.select_related(
            "owner",
            "eve_type",
            "eve_solar_system",
            "eve_solar_system__eve_constellation",
            "eve_solar_system__eve_constellation__eve_region",
            "poco_details",
            "eve_planet",
        ).filter(eve_type=EveTypeId.CUSTOMS_OFFICE, poco_details__isnull=False),
        id=structure_id,
    )
    context = {
        "structure": structure,
        "details": structure.poco_details,
        "last_updated": structure.last_updated_at,
    }
    return render(request, "structures/modals/poco_details.html", context)


@login_required
@permission_required("structures.basic_access")
def starbase_detail(request: HttpRequest, structure_id: int):
    """Shows detail modal for a starbase."""

    structure = get_object_or_404(
        Structure.objects.select_related(
            "owner",
            "owner__corporation",
            "owner__corporation__alliance",
            "eve_type",
            "eve_type__eve_group",
            "eve_solar_system",
            "eve_solar_system__eve_constellation",
            "eve_solar_system__eve_constellation__eve_region",
            "starbase_detail",
            "eve_moon",
        ).filter(starbase_detail__isnull=False),
        id=structure_id,
    )
    fuels = structure.starbase_detail.fuels.select_related("eve_type").order_by(
        "eve_type__name"
    )
    assets = defaultdict(int)
    for item in structure.items.select_related("eve_type"):
        assets[item.eve_type_id] += item.quantity
    eve_types: Dict[int, EveType] = EveType.objects.in_bulk(id_list=assets.keys())
    modules = sorted(
        [
            {"eve_type": eve_types.get(eve_type_id), "quantity": quantity}
            for eve_type_id, quantity in assets.items()
        ],
        key=lambda obj: obj["eve_type"].name,
    )
    modules_count = (
        functools.reduce(lambda x, y: x + y, [obj["quantity"] for obj in modules])
        if modules
        else 0
    )
    try:
        fuel_blocks_count = (
            structure.starbase_detail.fuels.filter(
                eve_type__eve_group_id=EveGroupId.FUEL_BLOCK
            )
            .first()
            .quantity
        )
    except AttributeError:
        fuel_blocks_count = None
    context = {
        "structure": structure,
        "detail": structure.starbase_detail,
        "fuels": fuels,
        "modules": modules,
        "modules_count": modules_count,
        "fuel_blocks_count": fuel_blocks_count,
        "last_updated_at": structure.last_updated_at,
    }
    return render(request, "structures/modals/starbase_detail.html", context)


@login_required
@permission_required("structures.add_structure_owner")
@token_required(scopes=Owner.get_esi_scopes())  # type: ignore
def add_structure_owner(request: HttpRequest, token: Token):
    """View for adding or replacing a structure owner."""
    token_char = get_object_or_404(EveCharacter, character_id=token.character_id)
    try:
        character_ownership = CharacterOwnership.objects.get(
            user=request.user, character=token_char
        )
    except CharacterOwnership.DoesNotExist:
        character_ownership = None
        messages.error(
            request,
            format_html(
                _(
                    "You can only use your main or alt characters "
                    "to add corporations. "
                    "However, character %s is neither. "
                )
                % token_char.character_name
            ),
        )
        return redirect("structures:index")
    try:
        corporation = EveCorporationInfo.objects.get(
            corporation_id=token_char.corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            token_char.corporation_id
        )
    owner, created = Owner.objects.update_or_create(
        corporation=corporation, defaults={"is_active": True}
    )
    owner.add_character(character_ownership)
    if created:
        default_webhooks = Webhook.objects.filter(is_default=True)
        if default_webhooks:
            for webhook in default_webhooks:
                owner.webhooks.add(webhook)
            owner.save()

    if owner.characters.count() == 1:
        tasks.update_all_for_owner.delay(owner_pk=owner.pk, user_pk=request.user.pk)  # type: ignore
        messages.info(
            request,
            format_html(
                _(
                    "%(corporation)s has been added with %(character)s "
                    "as sync character. "
                    "We have started fetching structures and notifications "
                    "for this corporation and you will receive a report once "
                    "the process is finished."
                )
                % {"corporation": owner, "character": token_char}
            ),
        )
        if STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                notify_admins(
                    message=_(
                        "%(corporation)s was added as new "
                        "structure owner by %(user)s."
                    )
                    % {"corporation": owner, "user": request.user.username},
                    title=_("%s: Structure owner added: %s") % (__title__, owner),
                )
    else:
        messages.info(
            request,
            format_html(
                _(
                    "%(character)s has been added to %(corporation)s "
                    "as sync character. "
                    "You now have %(characters_count)d sync character(s) configured."
                )
                % {
                    "corporation": owner,
                    "character": token_char,
                    "characters_count": owner.valid_characters_count(),
                }
            ),
        )
        if STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                notify_admins(
                    message=_(
                        "%(character)s was added as sync character to "
                        "%(corporation)s by %(user)s.\n"
                        "We now have %(characters_count)d sync character(s) configured."
                    )
                    % {
                        "character": token_char,
                        "corporation": owner,
                        "user": request.user.username,
                        "characters_count": owner.valid_characters_count(),
                    },
                    title=_("%s: Character added to: %s") % (__title__, owner),
                )
    return redirect("structures:index")



import datetime as dt
from unittest import mock
from unittest.mock import patch

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from structures.core.notification_types import NotificationType
from structures.models import GeneratedNotification, Structure
from structures.tests.testdata.factories import (
    GeneratedNotificationFactory,
    NotificationFactory,
    OwnerFactory,
    StarbaseFactory,
    StructureFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.notifications"


class TestGeneratedNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str(self):
        # given
        notif = GeneratedNotificationFactory()
        # when/then
        self.assertTrue(str(notif))

    def test_should_have_repr(self):
        # given
        notif = GeneratedNotificationFactory()
        # when/then
        self.assertTrue(repr(notif))

    def test_should_send_to_configured_webhooks(self):
        # given
        notif = GeneratedNotificationFactory()
        webhook = notif.owner.webhooks.first()
        webhook.notification_types = [NotificationType.TOWER_REINFORCED_EXTRA]
        webhook.save()
        # when
        with patch(MODULE_PATH + ".Webhook.send_message") as mock:
            mock.return_value = 1
            result = notif.send_to_configured_webhooks()
        # then
        self.assertTrue(result)


class TestGeneratedNotificationManagerCreatePosReinforced(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_should_create_new_notif(self):
        # given
        reinforced_until = now() + dt.timedelta(hours=24)
        starbase = StarbaseFactory(
            owner=self.owner,
            state=Structure.State.POS_REINFORCED,
            state_timer_end=reinforced_until,
        )
        # when
        obj, created = GeneratedNotification.objects.get_or_create_from_structure(
            structure=starbase, notif_type=NotificationType.TOWER_REINFORCED_EXTRA
        )
        # then
        self.assertTrue(created)
        self.assertEqual(
            dt.datetime.fromisoformat(obj.details["reinforced_until"]), reinforced_until
        )

    def test_should_return_existing_notif(self):
        # given
        reinforced_until = now() + dt.timedelta(hours=24)
        starbase = StarbaseFactory(
            owner=self.owner,
            state=Structure.State.POS_REINFORCED,
            state_timer_end=reinforced_until,
        )
        obj_old = GeneratedNotificationFactory(
            owner=starbase.owner,
            notif_type=NotificationType.TOWER_REINFORCED_EXTRA,
            details={"reinforced_until": reinforced_until.isoformat()},
            create_structure=False,
        )
        obj_old.structures.add(starbase)
        # when
        (
            obj_new,
            created,
        ) = GeneratedNotification.objects.get_or_create_from_structure(
            starbase, notif_type=NotificationType.TOWER_REINFORCED_EXTRA
        )
        # then
        self.assertFalse(created)
        self.assertEqual(obj_old, obj_new)

    def test_should_raise_error_when_no_starbase(self):
        # given
        reinforced_until = now() + dt.timedelta(hours=24)
        starbase = StructureFactory(state_timer_end=reinforced_until, owner=self.owner)
        # when
        with self.assertRaises(ValueError):
            GeneratedNotification.objects.get_or_create_from_structure(
                starbase, notif_type=NotificationType.TOWER_REINFORCED_EXTRA
            )

    def test_should_raise_error_when_not_reinforced(self):
        # given
        starbase = StarbaseFactory(owner=self.owner)
        # when
        with self.assertRaises(ValueError):
            GeneratedNotification.objects.get_or_create_from_structure(
                starbase, notif_type=NotificationType.TOWER_REINFORCED_EXTRA
            )

    def test_should_raise_error_when_reinforcement_timer_missing(self):
        # given
        starbase = StarbaseFactory(
            owner=self.owner, state=Structure.State.POS_REINFORCED
        )
        # when
        with self.assertRaises(ValueError):
            GeneratedNotification.objects.get_or_create_from_structure(
                starbase, notif_type=NotificationType.TOWER_REINFORCED_EXTRA
            )

    def test_should_raise_error_when_unsupported_notification_type(self):
        # given
        reinforced_until = now() + dt.timedelta(hours=24)
        structure = StarbaseFactory(
            owner=self.owner,
            state=Structure.State.POS_REINFORCED,
            state_timer_end=reinforced_until,
        )
        # when
        with self.assertRaises(ValueError):
            GeneratedNotification.objects.get_or_create_from_structure(
                structure, notif_type=NotificationType.TOWER_ALERT_MSG
            )


@mock.patch("structures.core.notification_timers.add_or_remove_timer")
class TestProcessTimers(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory()

    @mock.patch(MODULE_PATH + ".STRUCTURES_ADD_TIMERS", True)
    def test_should_process_timers_for_all_supported_notification_types(
        self, mock_add_or_remove_timer
    ):
        mock_add_or_remove_timer.return_value = True
        for notif_type in [
            NotificationType.STRUCTURE_LOST_SHIELD,
            NotificationType.STRUCTURE_LOST_ARMOR,
            NotificationType.ORBITAL_REINFORCED,
            NotificationType.MOONMINING_EXTRACTION_STARTED,
            NotificationType.MOONMINING_EXTRACTION_CANCELLED,
            NotificationType.SOV_STRUCTURE_REINFORCED,
            NotificationType.TOWER_REINFORCED_EXTRA,
        ]:
            with self.subTest(notif_type=notif_type):
                notif = NotificationFactory(owner=self.owner, notif_type=notif_type)
                self.assertTrue(notif.add_or_remove_timer())

    @mock.patch(MODULE_PATH + ".STRUCTURES_ADD_TIMERS", True)
    def test_should_not_process_timers_for_non_supported_notification_types(
        self, mock_add_or_remove_timer
    ):
        mock_add_or_remove_timer.return_value = True
        unsupported_notif_types = {
            obj for obj in NotificationType
        } - NotificationType.relevant_for_timerboard()
        for notif_type in unsupported_notif_types:
            with self.subTest(notif_type=notif_type):
                notif = NotificationFactory(owner=self.owner, notif_type=notif_type)
                self.assertFalse(notif.add_or_remove_timer())

    def test_should_not_process_timers_when_disabled(self, mock_add_or_remove_timer):
        # given
        mock_add_or_remove_timer.side_effect = RuntimeError
        notif = NotificationFactory(
            owner=self.owner, notif_type=NotificationType.CHAR_APP_ACCEPT_MSG
        )
        # when
        with mock.patch(MODULE_PATH + ".STRUCTURES_ADD_TIMERS", False):
            result = notif.add_or_remove_timer()
        # then
        self.assertFalse(result)
